"""Generic function to save and load python AT lattices. The format is
determined by the file extension
"""
import os.path
from at.lattice import Lattice

__all__ = ['load_lattice', 'save_lattice', 'register_format']

_load_extension = {}
_save_extension = {}


def load_lattice(filepath: str, **kwargs) -> Lattice:
    """Load a Lattice object from a file

The file format is indicated by the filepath extension.

Parameters:
    filepath:           Name of the file

Keyword Args:
    name (str):         Name of the lattice.
      Default: taken from the file, or ``''``
    energy (float):     Energy of the lattice
      (default: taken from the file)
    periodicity (int]): Number of periods
      (default: taken from the file, or 1)
    *:                  All other keywords will be set as :py:class:`.Lattice`
      attributes

Specific keywords for .mat files

Keyword Args:
    mat_key (str):      Name of the Matlab variable containing
      the lattice. Default: Matlab variable name if there is only one,
      otherwise ``'RING'``
    check (bool):       Run coherence tests. Default: :py:obj:`True`
    quiet (bool):       Suppress the warning for non-standard classes.
      Default: :py:obj:`False`
    keep_all (bool):    Keep Matlab RingParam elements as Markers.
      Default: :py:obj:`False`

Returns:
    lattice (Lattice):          New :py:class:`.Lattice` object

See Also:
    :py:func:`.load_mat`, :py:func:`.load_m`, :py:func:`.load_repr`,
    :py:func:`.load_elegant`, :py:func:`.load_tracy`

.. Admonition:: Known extensions are:
    """
    _, ext = os.path.splitext(filepath)
    try:
        load_func = _load_extension[ext.lower()]
    except KeyError:
        print("File load failed: unknow extension {}.".format(ext))
    else:
        return load_func(filepath, **kwargs)


def save_lattice(ring: Lattice, filepath: str, **kwargs):
    """Save a Lattice object

The file format is indicated by the filepath extension.

Parameters:
    ring:               Lattice description
    filepath:           Name of the file

Specific keywords for .mat files

Keyword Args:
    mat_key (str):      Name of the Matlab variable containing the lattice.
      Default: ``'RING'``

See Also:
    :py:func:`.save_mat`, :py:func:`.save_m`, :py:func:`.save_repr`

.. Admonition:: Known extensions are:
    """
    _, ext = os.path.splitext(filepath)
    try:
        save_func = _save_extension[ext.lower()]
    except KeyError:
        print("File save failed: unknow extension {}.".format(ext))
    else:
        return save_func(ring, filepath, **kwargs)


def register_format(extension: str, load_func=None, save_func=None,
                    descr: str = ''):
    """Register format-specific processing functions

    Parameters:
        extension:      File extension string.
        load_func:      load function. Default: :py:obj:`None`
        save_func:      save_lattice function Default: :py:obj:`None`
        descr:          File type description
    """
    if load_func is not None:
        _load_extension[extension] = load_func
        load_lattice.__doc__ += '\n    {0:<10}'\
                                '\n        {1}\n'.format(extension, descr)
    if save_func is not None:
        _save_extension[extension] = save_func
        save_lattice.__doc__ += '\n    {0:<10}'\
                                '\n        {1}\n'.format(extension, descr)


Lattice.load = staticmethod(load_lattice)
Lattice.save = save_lattice



"""
Collection of functions to fit various global parameters
such as the tune and the chromaticity
"""
from typing import Optional
import numpy
from ..lattice import Lattice, Refpts, set_value_refpts
from ..lattice import AtWarning, AtError
from ..physics import get_tune, get_chrom


__all__ = ['fit_tune', 'fit_chrom']


def _get_tune(ring: Lattice, dp: float, **kwargs):
    get_integer = kwargs.pop('fit_integer', False)
    return get_tune(ring, dp=dp, get_integer=get_integer)[0:2]


def _get_chrom(ring: Lattice, dp: float, **kwargs):
    return get_chrom(ring, dp=dp)[0:2]


def _fit_tune_chrom(ring: Lattice, index: int, func,
                    refpts1: Refpts, refpts2: Refpts, newval,
                    tol: Optional[float] = 1.0e-12,
                    dp: Optional[float] = 0, niter: Optional[int] = 3,
                    **kwargs):

    def _get_resp(ring: Lattice, index: int, func, refpts, attname,
                  delta, dp, **kwargs):
        set_value_refpts(ring, refpts, attname, delta, index=index,
                         increment=True)
        datap = func(ring, dp, **kwargs)
        set_value_refpts(ring, refpts, attname, -2*delta, index=index,
                         increment=True)
        datan = func(ring, dp, **kwargs)
        set_value_refpts(ring, refpts, attname, delta, index=index,
                         increment=True)
        data = numpy.subtract(datap, datan)/(2*delta)
        return data

    def _fit(ring, index, func, refpts1, refpts2, newval, J,
             dp: Optional[float] = 0, **kwargs):
        val = func(ring, dp, **kwargs)
        dk = numpy.linalg.solve(J, numpy.subtract(newval, val))
        set_value_refpts(ring, refpts1, 'PolynomB', dk[0], index=index,
                         increment=True)
        set_value_refpts(ring, refpts2, 'PolynomB', dk[1], index=index,
                         increment=True)
        val = func(ring, dp, **kwargs)
        sumsq = numpy.sum(numpy.square(numpy.subtract(val, newval)))
        return sumsq

    delta = 1e-6 * 10 ** index
    dq1 = _get_resp(ring, index, func, refpts1, 'PolynomB',
                    delta, dp, **kwargs)
    dq2 = _get_resp(ring, index, func, refpts2, 'PolynomB',
                    delta, dp, **kwargs)
    J = [[dq1[0], dq2[0]], [dq1[1], dq2[1]]]

    n = 0
    sumsq = tol+1
    print('Initial value', func(ring, dp, **kwargs))
    while sumsq > tol and n < niter:
        sumsq = _fit(ring, index, func, refpts1, refpts2, newval,
                     J, dp=dp, **kwargs)
        print('iter#', n, 'Res.', sumsq)
        n += 1
    print('Final value', func(ring, dp, **kwargs), '\n')
    return


def fit_tune(ring: Lattice, refpts1: Refpts, refpts2: Refpts, newval,
             tol: float = 1.0e-12,
             dp: Optional[float] = 0, niter: int = 3,
             **kwargs) -> None:
    """Fits the tunes using 2 families

    Args:
        ring:       Lattice description
        refpts1:    Selection of the 1st family
        refpts2:    Selection of the 2nd family
        newval:     New tunes, in case an non-zero integer part
                    is provided, fit_integer is set to True
                    
    Keyword arguments:
        tol:        Tolerance for the matching; Default: 1.0e-12
        dp:         Momentum deviation. Default: 0
        niter:      Maximum number of iterations. Default 3
        fit_integer: bool (default=False), use integer tune

    Typical usage:
    at.fit_tune(ring, refpts1, refpts2, [0.1,0.25])
    """
    print('\nFitting Tune...')
    if numpy.any(numpy.floor(newval) != 0.0):
        kwargs['fit_integer'] = True
    _fit_tune_chrom(ring, 1, _get_tune, refpts1, refpts2, newval, tol=tol,
                    dp=dp, niter=niter, **kwargs)


def fit_chrom(ring: Lattice, refpts1: Refpts, refpts2: Refpts, newval,
              tol: Optional[float] = 1.0e-12,
              dp: Optional[float] = 0, niter: Optional[int] = 3,
              **kwargs) -> None:
    """Fit the chromaticities using 2 families

    Args:
        ring:       Lattice description
        refpts1:    Selection of the 1st family
        refpts2:    Selection of the 2nd family
        newval:     New tunes
        tol:        Tolerance for the matching; Default: 1.0e-12
        dp:         Momentum deviation. Default: 0
        niter:      Maximum number of iterations. Default 3

    Typical usage:
    at.fit_chrom(ring, refpts1, refpts2, [10,5])
    """
    print('\nFitting Chromaticity...')
    _fit_tune_chrom(ring, 2, _get_chrom, refpts1, refpts2, newval, tol=tol,
                    dp=dp, niter=niter)



"""
Module to define common elements used in AT.

Each element has a default PassMethod attribute for which it should have the
appropriate attributes.  If a different PassMethod is set, it is the caller's
responsibility to ensure that the appropriate attributes are present.
"""
import abc
import re
import numpy
from copy import copy, deepcopy
from abc import ABC
from typing import Optional, Generator, Tuple, List, Iterable


def _array(value, shape=(-1,), dtype=numpy.float64):
    # Ensure proper ordering(F) and alignment(A) for "C" access in integrators
    return numpy.require(value, dtype=dtype, requirements=['F', 'A']).reshape(
        shape, order='F')


def _array66(value):
    return _array(value, shape=(6, 6))


def _nop(value):
    return value


class LongtMotion(ABC):
    """Abstract Base class for all Element classes whose instances may modify
    the particle momentum

    Allows to identify elements potentially inducing longitudinal motion.

    Subclasses of :py:class:`LongtMotion` must provide two methods for
    enabling longitudinal motion:

    * ``_get_longt_motion(self)`` must return the activation state,
    * ``set_longt_motion(self, enable, new_pass=None, copy=False, **kwargs)``
      must enable or disable longitudinal motion.
    """
    @abc.abstractmethod
    def _get_longt_motion(self):
        return False

    # noinspection PyShadowingNames
    @abc.abstractmethod
    def set_longt_motion(self, enable, new_pass=None, copy=False, **kwargs):
        """Enable/Disable longitudinal motion

        Parameters:
            enable:     ``True`` for enabling, ``False`` for disabling
            new_pass:   New PassMethod:

              * ``None``: makes no change,
              * ``'auto'``: Uses the default conversion,
              * Anything else is used as the new PassMethod.
            copy:       If True, returns a modified copy of the element,
              otherwise modifies the element in-place
        """
        # noinspection PyUnresolvedReferences
        if new_pass is None or new_pass == self.PassMethod:
            return self if copy else None
        if copy:
            newelem = deepcopy(self)
            newelem.PassMethod = new_pass
            return newelem
        # noinspection PyAttributeOutsideInit
        self.PassMethod = new_pass


# noinspection PyUnresolvedReferences
class _DictLongtMotion(LongtMotion):
    # noinspection PyShadowingNames
    """Mixin class for elements implementing a 'default_pass' class attribute

    :py:class:`DictLongtMotion` provides:

    * a :py:meth:`set_longt_motion` method setting the PassMethod according
      to the ``default_pass`` dictionary.
    * a :py:obj:`.longt_motion` property set to ``True`` when the PassMethod
      is ``default_pass[True]``

    The class must have a ``default_pass`` class attribute, a dictionary
    such that:

    * ``default_pass[False]`` is the PassMethod when radiation is turned
      OFF,
    * ``default_pass[True]`` is the default PassMethod when radiation is
      turned ON.

    The :py:class:`DictLongtMotion` class must be set as the first base class.

    Example:

        >>> class QuantumDiffusion(_DictLongtMotion, Element):
        ...
        ...     default_pass = {False: 'IdentityPass', True: 'QuantDiffPass'}

        Defines a class such that :py:meth:`set_longt_motion` will select
        ``'IdentityPass'`` or ``'IdentityPass'``.
        """
    def _get_longt_motion(self):
        return self.PassMethod != self.default_pass[False]

    # noinspection PyShadowingNames
    def set_longt_motion(self, enable, new_pass=None, **kwargs):
        if new_pass == 'auto':
            new_pass = self.default_pass[enable]
        return super().set_longt_motion(enable, new_pass=new_pass, **kwargs)


# noinspection PyUnresolvedReferences
class _Radiative(LongtMotion):
    # noinspection PyShadowingNames
    r"""Mixin class for radiating elements

    :py:class:`_Radiative` implements the mechanism for converting the pass
    methods of radiating elements. It provides:

    * a :py:meth:`set_longt_motion` method setting the PassMethod
      according to the following rule:

      * ``enable == True``: replace "\*Pass" by "\*RadPass"
      * ``enable == False``: replace "\*RadPass" by "\*Pass"
    * a :py:obj:`.longt_motion` property set to true when the PassMethod
      ends with "RadPass"

    The :py:class:`_Radiative` class must be set as the first base class.

    Example:
        >>> class Multipole(_Radiative, LongElement, ThinMultipole):

        Defines a class where :py:meth:`set_longt_motion` will convert the
        PassMethod according to the \*Pass or \*RadPass suffix.
    """

    def _get_longt_motion(self):
        return self.PassMethod.endswith('RadPass')

    def _autopass(self, enable):
        rad = self.longt_motion
        if enable and not rad:
            return ''.join((self.PassMethod[:-4], 'RadPass'))
        elif not enable and rad:
            return ''.join((self.PassMethod[:-7], 'Pass'))
        else:
            return None

    # noinspection PyTypeChecker,PyShadowingNames
    def set_longt_motion(self, enable, new_pass=None, copy=False, **kwargs):
        if new_pass == 'auto':
            new_pass = self._autopass(enable)
        if new_pass is None or new_pass == self.PassMethod:
            return self if copy else None
        if enable:
            def setpass(el):
                el.PassMethod = new_pass
                el.Energy = kwargs['energy']
        else:
            def setpass(el):
                el.PassMethod = new_pass
                try:
                    del el.Energy
                except AttributeError:
                    pass
        if copy:
            newelem = deepcopy(self)
            setpass(newelem)
            return newelem
        setpass(self)


class Radiative(_Radiative):
    # noinspection PyShadowingNames
    r"""Mixin class for default radiating elements (:py:class:`.Dipole`,
    :py:class:`.Quadrupole`, :py:class:`.Wiggler`)

    :py:class:`Radiative` is a base class for the subset of radiative elements
    considered as the ones to be turned on by default: :py:class:`.Dipole`,
    :py:class:`.Quadrupole` and :py:class:`.Wiggler`, excluding the higher
    order multipoles.

    :py:class:`Radiative` inherits from :py:class:`_Radiative` and does not
    add any new functionality. Its purpose is to identify the default set of
    radiating elements.

    Example:
        >>> class Dipole(Radiative, Multipole):

        Defines a class belonging to the default radiating elements. It
        converts the PassMethod according to the "\*Pass" or "\*RadPass"
        suffix.
    """
    pass


class Collective(_DictLongtMotion):
    """Mixin class for elements representing collective effects

    Derived classes will automatically set the :py:obj:`is_collective`
    property when the element is active.

    The class must have a ``default_pass`` class attribute, a dictionary such
    that:

    * ``default_pass[False]`` is the PassMethod when collective effects
      are turned OFF,
    * ``default_pass[True]`` is the default PassMethod when collective effects
      are turned ON.

    The :py:class:`Collective` class must be set as the first base class.

    Example:
        >>> class WakeElement(Collective, Element):
        ...
        ...     default_pass = {False: 'IdentityPass', True: 'WakeFieldPass'}

        Defines a class where the :py:obj:`is_collective` property is handled
    """

    def _get_collective(self):
        # noinspection PyUnresolvedReferences
        return self.PassMethod != self.default_pass[False]


class Element(object):
    """Base class for AT elements"""

    _BUILD_ATTRIBUTES = ['FamName']
    _conversions = dict(FamName=str, PassMethod=str, Length=float,
                        R1=_array66, R2=_array66,
                        T1=lambda v: _array(v, (6,)),
                        T2=lambda v: _array(v, (6,)),
                        RApertures=lambda v: _array(v, (4,)),
                        EApertures=lambda v: _array(v, (2,)),
                        KickAngle=lambda v: _array(v, (2,)),
                        PolynomB=_array, PolynomA=_array,
                        BendingAngle=float,
                        MaxOrder=int, NumIntSteps=int,
                        Energy=float,
                        )

    _entrance_fields = ['T1', 'R1']
    _exit_fields = ['T2', 'R2']

    def __init__(self, family_name: str, **kwargs):
        """
        Parameters:
            family_name:    Name of the element

        All keywords will be set as attributes of the element
        """

        self.FamName = family_name
        self.Length = kwargs.pop('Length', 0.0)
        self.PassMethod = kwargs.pop('PassMethod', 'IdentityPass')
        self.update(kwargs)

    def __setattr__(self, key, value):
        try:
            super(Element, self).__setattr__(
                key, self._conversions.get(key, _nop)(value))
        except Exception as exc:
            exc.args = ('In element {0}, parameter {1}: {2}'.format(
                self.FamName, key, exc),)
            raise

    def __str__(self):
        first3 = ['FamName', 'Length', 'PassMethod']
        attrs = dict(self.items())
        keywords = ['\t{0} : {1!s}'.format(k, attrs.pop(k)) for k in first3]
        keywords += ['\t{0} : {1!s}'.format(k, v) for k, v in attrs.items()]
        return '\n'.join((type(self).__name__ + ':', '\n'.join(keywords)))

    def __repr__(self):
        attrs = dict(self.items())
        arguments = [attrs.pop(k, getattr(self, k)) for k in
                     self._BUILD_ATTRIBUTES]
        defelem = self.__class__(*arguments)
        keywords = ['{0!r}'.format(arg) for arg in arguments]
        keywords += ['{0}={1!r}'.format(k, v) for k, v in sorted(attrs.items())
                     if not numpy.array_equal(v, getattr(defelem, k, None))]
        args = re.sub(r'\n\s*', ' ', ', '.join(keywords))
        return '{0}({1})'.format(self.__class__.__name__, args)

    def equals(self, other) -> bool:
        """Whether an element is equivalent to another.

        This implementation was found to be too slow for the generic
        __eq__ method when comparing lattices.
        """
        return repr(self) == repr(other)

    def divide(self, frac) -> List["Element"]:
        """split the element in len(frac) pieces whose length
        is frac[i]*self.Length

        Parameters:
            frac:           length of each slice expressed as a fraction of the
              initial length. ``sum(frac)`` may differ from 1.

        Returns:
            elem_list:  a list of elements equivalent to the original.

        Example:

            >>> Drift('dr', 0.5).divide([0.2, 0.6, 0.2])
            [Drift('dr', 0.1), Drift('dr', 0.3), Drift('dr', 0.1)]
        """
        # Bx default, the element is indivisible
        return [self]

    def update(self, *args, **kwargs):
        """
        update(**kwargs)
        update(mapping, **kwargs)
        update(iterable, **kwargs)
        Update the element attributes with the given arguments
        """
        attrs = dict(*args, **kwargs)
        for (key, value) in attrs.items():
            setattr(self, key, value)

    def copy(self) -> "Element":
        """Return a shallow copy of the element"""
        return copy(self)

    def deepcopy(self) -> "Element":
        """Return a deep copy of the element"""
        return deepcopy(self)

    def items(self) -> Generator[Tuple, None, None]:
        """Iterates through the data members"""
        for k, v in vars(self).items():
            yield k, v

    def is_compatible(self, other) -> bool:
        """Checks if another Element can be merged"""
        return False

    def merge(self, other) -> None:
        """Merge another element"""
        if not self.is_compatible(other):
            badname = getattr(other, 'FamName', type(other))
            raise TypeError('Cannot merge {0} and {1}'.format(self.FamName,
                                                              badname))

    # noinspection PyMethodMayBeStatic
    def _get_longt_motion(self):
        return False

    # noinspection PyMethodMayBeStatic
    def _get_collective(self):
        return False

    @property
    def longt_motion(self):
        """``True`` if longitudinal motion is affected by the element"""
        return self._get_longt_motion()

    @property
    def is_collective(self):
        """``True`` if the element involves collective effects"""
        return self._get_collective()


class LongElement(Element):
    """Base class for long elements
    """
    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ['Length']

    def __init__(self, family_name: str, length: float, *args, **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]

        Other arguments and keywords are given to the base class
        """
        kwargs.setdefault('Length', length)
        # Ancestor may be either Element of ThinMultipole
        # noinspection PyArgumentList
        super(LongElement, self).__init__(family_name, *args, **kwargs)

    def _part(self, fr, sumfr):
        pp = self.copy()
        pp.Length = fr * self.Length
        if hasattr(self, 'KickAngle'):
            pp.KickAngle = fr / sumfr * self.KickAngle
        return pp

    def divide(self, frac) -> List[Element]:
        def popattr(element, attr):
            val = getattr(element, attr)
            delattr(element, attr)
            return attr, val

        frac = numpy.asarray(frac, dtype=float)
        el = self.copy()
        # Remove entrance and exit attributes
        fin = dict(popattr(el, key) for key in vars(self) if
                   key in self._entrance_fields)
        fout = dict(popattr(el, key) for key in vars(self) if
                    key in self._exit_fields)
        # Split element
        element_list = [el._part(f, numpy.sum(frac)) for f in frac]
        # Restore entrance and exit attributes
        for key, value in fin.items():
            setattr(element_list[0], key, value)
        for key, value in fout.items():
            setattr(element_list[-1], key, value)
        return element_list

    def is_compatible(self, other) -> bool:
        return type(other) is type(self) and \
               self.PassMethod == other.PassMethod

    def merge(self, other) -> None:
        super().merge(other)
        self.Length += other.Length


class Marker(Element):
    """Marker element"""


class Monitor(Element):
    """Monitor element"""


class Aperture(Element):
    """Aperture element"""
    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ['Limits']
    _conversions = dict(Element._conversions, Limits=lambda v: _array(v, (4,)))

    def __init__(self, family_name, limits, **kwargs):
        """
        Args:
            family_name:    Name of the element
            limits:         (4,) array of physical aperture:
              [xmin, xmax, zmin, zmax] [m]

        Default PassMethod: ``AperturePass``
        """
        kwargs.setdefault('PassMethod', 'AperturePass')
        super(Aperture, self).__init__(family_name, Limits=limits, **kwargs)


class Drift(LongElement):
    """Drift space element"""

    def __init__(self, family_name: str, length: float, **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]

        Default PassMethod: ``DriftPass``
        """
        kwargs.setdefault('PassMethod', 'DriftPass')
        super(Drift, self).__init__(family_name, length, **kwargs)

    def insert(self,
               insert_list: Iterable[Tuple[float, Optional[Element]]]) \
            -> List[Element]:
        """insert elements inside a drift

        Arguments:
            insert_list: iterable, each item of insert_list is itself an
              iterable with 2 objects:

              1. the location where the center of the element
                 will be inserted, given as a fraction of the Drift length.
              2. an element to be inserted at that location. If ``None``,
                 the drift will be divided but no element will be inserted.

        Returns:
             elem_list: a list of elements.

        Drifts with negative lengths may be generated if necessary.

        Examples:

            >>> Drift('dr', 2.0).insert(((0.25, None), (0.75, None)))
            [Drift('dr', 0.5), Drift('dr', 1.0), Drift('dr', 0.5)]

            >>> Drift('dr', 2.0).insert(((0.0, Marker('m1')),
            ... (0.5, Marker('m2'))))
            [Marker('m1'), Drift('dr', 1.0), Marker('m2'), Drift('dr', 1.0)]

            >>> Drift('dr', 2.0).insert(((0.5, Quadrupole('qp', 0.4, 0.0)),))
            [Drift('dr', 0.8), Quadrupole('qp', 0.4), Drift('dr', 0.8)]
        """
        frac, elements = zip(*insert_list)
        lg = [0.0 if el is None else el.Length for el in elements]
        fr = numpy.asarray(frac, dtype=float)
        lg = 0.5 * numpy.asarray(lg, dtype=float) / self.Length
        drfrac = numpy.hstack((fr - lg, 1.0)) - numpy.hstack((0.0, fr + lg))
        long_elems = (drfrac != 0.0)
        drifts = numpy.ndarray((len(drfrac),), dtype='O')
        drifts[long_elems] = self.divide(drfrac[long_elems])
        nline = len(drifts) + len(elements)
        line = [None] * nline  # type: List[Optional[Element]]
        line[::2] = drifts
        line[1::2] = elements
        return [el for el in line if el is not None]


class Collimator(Drift):
    """Collimator element"""
    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['RApertures']

    def __init__(self, family_name: str, length: float, limits, **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            limits:         (4,) array of physical aperture:
              [xmin, xmax, zmin, zmax] [m]

        Default PassMethod: ``DriftPass``
        """
        super(Collimator, self).__init__(family_name, length,
                                         RApertures=limits, **kwargs)


class ThinMultipole(Element):
    """Thin multipole element"""
    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ['PolynomA',
                                                     'PolynomB']

    def __init__(self, family_name: str, poly_a, poly_b, **kwargs):
        """
        Args:
            family_name:    Name of the element
            poly_a:         Array of normal multipole components
            poly_b:         Array of skew multipole components

        Keyword arguments:
            MaxOrder:   Number of desired multipoles. Default: highest index of
                        non-zero polynomial coefficients

        Default PassMethod: ``ThinMPolePass``
        """

        def getpol(poly):
            nonzero = numpy.flatnonzero(poly != 0.0)
            return poly, len(poly), nonzero[-1] if len(nonzero) > 0 else -1

        def lengthen(poly, dl):
            if dl > 0:
                return numpy.concatenate((poly, numpy.zeros(dl)))
            else:
                return poly

        # Remove MaxOrder, PolynomA and PolynomB
        poly_a, len_a, ord_a = getpol(_array(kwargs.pop('PolynomA', poly_a)))
        poly_b, len_b, ord_b = getpol(_array(kwargs.pop('PolynomB', poly_b)))
        deforder = max(getattr(self, 'DefaultOrder', 0), ord_a, ord_b)
        maxorder = kwargs.pop('MaxOrder', deforder)
        kwargs.setdefault('PassMethod', 'ThinMPolePass')
        super(ThinMultipole, self).__init__(family_name, **kwargs)
        # Set MaxOrder while PolynomA and PolynomB are not set yet
        super(ThinMultipole, self).__setattr__('MaxOrder', maxorder)
        # Adjust polynom lengths and set them
        len_ab = max(self.MaxOrder + 1, len_a, len_b)
        self.PolynomA = lengthen(poly_a, len_ab - len_a)
        self.PolynomB = lengthen(poly_b, len_ab - len_b)

    def __setattr__(self, key, value):
        """Check the compatibility of MaxOrder, PolynomA and PolynomB"""
        polys = ('PolynomA', 'PolynomB')
        if key in polys:
            value = _array(value)
            lmin = getattr(self, 'MaxOrder')
            if not len(value) > lmin:
                raise ValueError(
                    'Length of {0} must be larger than {1}'.format(key, lmin))
        elif key == 'MaxOrder':
            value = int(value)
            lmax = min(len(getattr(self, k)) for k in polys)
            if not value < lmax:
                raise ValueError(
                    'MaxOrder must be smaller than {0}'.format(lmax))

        super(ThinMultipole, self).__setattr__(key, value)


class Multipole(_Radiative, LongElement, ThinMultipole):
    """Multipole element"""
    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['PolynomA',
                                                         'PolynomB']
    _conversions = dict(ThinMultipole._conversions, K=float, H=float)

    def __init__(self, family_name: str, length: float, poly_a, poly_b,
                 **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            poly_a:         Array of normal multipole components
            poly_b:         Array of skew multipole components

        Keyword arguments:
            MaxOrder:   Number of desired multipoles. Default: highest index of
              non-zero polynomial coefficients
            NumIntSteps: Number of integration steps (default: 10)
            KickAngle:  Correction deviation angles (H, V)

        Default PassMethod: ``StrMPoleSymplectic4Pass``
        """
        kwargs.setdefault('PassMethod', 'StrMPoleSymplectic4Pass')
        kwargs.setdefault('NumIntSteps', 10)
        super(Multipole, self).__init__(family_name, length,
                                        poly_a, poly_b, **kwargs)

    def is_compatible(self, other) -> bool:
        if super().is_compatible(other) and \
                self.MaxOrder == other.MaxOrder:
            for i in range(self.MaxOrder + 1):
                if self.PolynomB[i] != other.PolynomB[i]:
                    return False
                if self.PolynomA[i] != other.PolynomA[i]:
                    return False
            return True
        else:
            return False

    # noinspection PyPep8Naming
    @property
    def K(self) -> float:
        """Focusing strength [mˆ-2]"""
        return 0.0 if len(self.PolynomB) < 2 else self.PolynomB[1]

    # noinspection PyPep8Naming
    @K.setter
    def K(self, strength: float):
        self.PolynomB[1] = strength

    # noinspection PyPep8Naming
    @property
    def H(self) -> float:
        """Sextupolar strength"""
        return 0.0 if len(self.PolynomB) < 3 else self.PolynomB[2]

    # noinspection PyPep8Naming
    @H.setter
    def H(self, strength):
        self.PolynomB[2] = strength


class Dipole(Radiative, Multipole):
    """Dipole element"""
    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['BendingAngle',
                                                         'K']
    _conversions = dict(Multipole._conversions, EntranceAngle=float,
                        ExitAngle=float,
                        FringeInt1=float, FringeInt2=float,
                        FringeQuadEntrance=int, FringeQuadExit=int,
                        FringeBendEntrance=int, FringeBendExit=int)

    _entrance_fields = Multipole._entrance_fields + ['EntranceAngle',
                                                     'FringeInt1',
                                                     'FringeBendEntrance'
                                                     'FringeQuadEntrance']
    _exit_fields = Multipole._exit_fields + ['ExitAngle',
                                             'FringeInt2',
                                             'FringeBendExit',
                                             'FringeQuadExit']

    DefaultOrder = 0

    def __init__(self, family_name: str, length: float,
                 bending_angle: Optional[float] = 0.0, k: float = 0.0,
                 **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            bending_angle:  Bending angle [rd]
            poly_a:         Array of normal multipole components
            poly_b:         Array of skew multipole components
            k=0:            Field index

        Keyword arguments:
            EntranceAngle=0.0:  entrance angle
            ExitAngle=0.0:      exit angle
            PolynomB:           straight multipoles
            PolynomA:           skew multipoles
            MaxOrder:           Number of desired multipoles
            NumIntSt=10:        Number of integration steps
            FullGap:            Magnet full gap
            FringeInt1:         Fringe field extension
            FringeInt2:
            FringeBendEntrance: 1: legacy version Brown First Order (default)

              2: SOLEIL close to second order of Brown

              3: THOMX
            FringeBendExit:     See ``FringeBendEntrance``
            FringeQuadEntrance: 0: no fringe field effect (default)

              1: Lee-Whiting's thin lens limit formula

              2: elegant-like
            FringeQuadExit:     See ``FringeQuadEntrance``
            fringeIntM0:        Integrals for FringeQuad method 2
            fringeIntP0:
            KickAngle:          Correction deviation angles (H, V)

        Default PassMethod: ``BndMPoleSymplectic4Pass``
        """
        poly_b = kwargs.pop('PolynomB', numpy.array([0, k]))
        kwargs.setdefault('BendingAngle', bending_angle)
        kwargs.setdefault('EntranceAngle', 0.0)
        kwargs.setdefault('ExitAngle', 0.0)
        kwargs.setdefault('PassMethod', 'BndMPoleSymplectic4Pass')
        super(Dipole, self).__init__(family_name, length, [], poly_b, **kwargs)

    def items(self) -> Generator[Tuple, None, None]:
        yield from super().items()
        yield 'K', self.K

    def _part(self, fr, sumfr):
        pp = super(Dipole, self)._part(fr, sumfr)
        pp.BendingAngle = fr / sumfr * self.BendingAngle
        pp.EntranceAngle = 0.0
        pp.ExitAngle = 0.0
        return pp

    def is_compatible(self, other) -> bool:
        def invrho(dip: Dipole):
            return dip.BendingAngle / dip.Length

        return (super().is_compatible(other) and
                self.ExitAngle == -other.EntranceAngle and
                abs(invrho(self) - invrho(other)) <= 1.e-6)

    def merge(self, other) -> None:
        super().merge(other)
        # noinspection PyAttributeOutsideInit
        self.ExitAngle = other.ExitAngle
        self.BendingAngle += other.BendingAngle


# Bend is a synonym of Dipole.
Bend = Dipole


class Quadrupole(Radiative, Multipole):
    """Quadrupole element"""
    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['K']
    _conversions = dict(Multipole._conversions, FringeQuadEntrance=int,
                        FringeQuadExit=int)

    _entrance_fields = Multipole._entrance_fields + ['FringeQuadEntrance']
    _exit_fields = Multipole._exit_fields + ['FringeQuadExit']

    DefaultOrder = 1

    def __init__(self, family_name: str, length: float,
                 k: Optional[float] = 0.0, **kwargs):
        """Quadrupole(FamName, Length, Strength=0, **keywords)

        Args:
            family_name:    Name of the element
            length:         Element length [m]
            k:              strength [mˆ-2]

        Keyword Arguments:
            PolynomB:           straight multipoles
            PolynomA:           skew multipoles
            MaxOrder:           Number of desired multipoles
            NumIntSteps=10:     Number of integration steps
            FringeQuadEntrance: 0: no fringe field effect (default)

              1: Lee-Whiting's thin lens limit formula

              2: elegant-like
            FringeQuadExit:     See ``FringeQuadEntrance``
            fringeIntM0:        Integrals for FringeQuad method 2
            fringeIntP0:
            KickAngle:          Correction deviation angles (H, V)

        Default PassMethod: ``StrMPoleSymplectic4Pass``
        """
        poly_b = kwargs.pop('PolynomB', numpy.array([0, k]))
        kwargs.setdefault('PassMethod', 'StrMPoleSymplectic4Pass')
        super(Quadrupole, self).__init__(family_name, length, [], poly_b,
                                         **kwargs)

    def items(self) -> Generator[Tuple, None, None]:
        yield from super().items()
        yield 'K', self.K


class Sextupole(Multipole):
    """Sextupole element"""
    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['H']

    DefaultOrder = 2

    def __init__(self, family_name: str, length: float,
                 h: Optional[float] = 0.0, **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            h:              strength [mˆ-3]

        Keyword Arguments:
            PolynomB:           straight multipoles
            PolynomA:           skew multipoles
            MaxOrder:           Number of desired multipoles
            NumIntSteps=10:     Number of integration steps
            KickAngle:          Correction deviation angles (H, V)

        Default PassMethod: ``StrMPoleSymplectic4Pass``
        """
        poly_b = kwargs.pop('PolynomB', [0, 0, h])
        kwargs.setdefault('PassMethod', 'StrMPoleSymplectic4Pass')
        super(Sextupole, self).__init__(family_name, length, [], poly_b,
                                        **kwargs)

    def items(self) -> Generator[Tuple, None, None]:
        yield from super().items()
        yield 'H', self.H


class Octupole(Multipole):
    """Octupole element, with no changes from multipole at present"""
    _BUILD_ATTRIBUTES = Multipole._BUILD_ATTRIBUTES

    DefaultOrder = 3


class RFCavity(LongtMotion, LongElement):
    """RF cavity element"""
    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['Voltage',
                                                         'Frequency',
                                                         'HarmNumber',
                                                         'Energy']
    default_pass = {False: 'DriftPass', True: 'RFCavityPass'}
    _conversions = dict(LongElement._conversions,
                        Voltage=float, Frequency=float,
                        HarmNumber=int, TimeLag=float)

    def __init__(self, family_name: str, length: float, voltage: float,
                 frequency: float, harmonic_number: int, energy: float,
                 **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            voltage:        RF voltage [V]
            frequency:      RF frequency [Hz]
            harmonic_number:
            energy:         ring energy [eV]

        Keyword Arguments:
            TimeLag=0:      Cavity time lag

        Default PassMethod: ``RFCavityPass``
        """
        kwargs.setdefault('TimeLag', 0.0)
        kwargs.setdefault('PassMethod', self.default_pass[True])
        super(RFCavity, self).__init__(family_name, length,
                                       Voltage=voltage,
                                       Frequency=frequency,
                                       HarmNumber=harmonic_number,
                                       Energy=energy, **kwargs)

    def _part(self, fr, sumfr):
        pp = super(RFCavity, self)._part(fr, sumfr)
        pp.Voltage = fr * self.Voltage
        return pp

    def is_compatible(self, other) -> bool:
        return (super().is_compatible(other) and
                self.Frequency == other.Frequency and
                self.TimeLag == other.TimeLag)

    def merge(self, other) -> None:
        super().merge(other)
        self.Voltage += other.Voltage

    def _get_longt_motion(self):
        return self.PassMethod.endswith('CavityPass')

    # noinspection PyShadowingNames
    def set_longt_motion(self, enable, new_pass=None, **kwargs):
        if new_pass == 'auto':
            new_pass = (self.default_pass[True] if enable else
                        ('IdentityPass' if self.Length == 0 else 'DriftPass'))
        return super().set_longt_motion(enable, new_pass=new_pass, **kwargs)


class M66(Element):
    """Linear (6, 6) transfer matrix"""
    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES
    _conversions = dict(Element._conversions, M66=_array66)

    def __init__(self, family_name: str, m66=None, **kwargs):
        """
        Args:
            family_name:    Name of the element
            m66:            Transfer matrix. Default: Identity matrix

        Default PassMethod: ``Matrix66Pass``
       """
        if m66 is None:
            m66 = numpy.identity(6)
        kwargs.setdefault('PassMethod', 'Matrix66Pass')
        super(M66, self).__init__(family_name, M66=m66, **kwargs)


class Corrector(LongElement):
    """Corrector element"""
    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['KickAngle']

    def __init__(self, family_name: str, length: float, kick_angle, **kwargs):
        """
        Args:
            family_name:    Name of the element
            length:         Element length [m]
            KickAngle:      Correction deviation angles (H, V)

        Default PassMethod: ``CorrectorPass``
        """
        kwargs.setdefault('PassMethod', 'CorrectorPass')
        super(Corrector, self).__init__(family_name, length,
                                        KickAngle=kick_angle, **kwargs)


class Wiggler(Radiative, LongElement):
    """Wiggler element

    See atwiggler.m
    """
    _BUILD_ATTRIBUTES = LongElement._BUILD_ATTRIBUTES + ['Lw', 'Bmax',
                                                         'Energy']
    _conversions = dict(Element._conversions, Lw=float, Bmax=float,
                        Energy=float,
                        Bx=lambda v: _array(v, (6, -1)),
                        By=lambda v: _array(v, (6, -1)),
                        Nstep=int, Nmeth=int, NHharm=int, NVharm=int)

    # noinspection PyPep8Naming
    def __init__(self, family_name: str, length: float, wiggle_period: float,
                 b_max: float, energy: float, Nstep: Optional[int] = 5,
                 Nmeth: Optional[int] = 4,
                 By=(1, 1, 0, 1, 1, 0), Bx=(), **kwargs):
        """
        Args:
            length:         total length of the wiggler
            wiggle_period:  length must be a multiple of this
            b_max:          peak wiggler field [Tesla]
            energy:         beam energy [eV]
            Nstep:          number of integration steps.
            Nmeth:          symplectic integration order: 2 or 4
            Bx:             harmonics for horizontal wiggler: (6, nHharm)
                              array-like object
            By:             harmonics for vertical wiggler (6, nHharm)
                              array-like object

        Default PassMethod: ``GWigSymplecticPass``
        """
        kwargs.setdefault('PassMethod', 'GWigSymplecticPass')
        n_wiggles = length / wiggle_period
        if abs(round(n_wiggles) - n_wiggles) > 1e-6:
            raise ValueError("Wiggler: length / wiggle_period is not an "
                             "integer. ({0}/{1}={2})".format(length,
                                                             wiggle_period,
                                                             n_wiggles))
        super(Wiggler, self).__init__(family_name, length, Lw=wiggle_period,
                                      Bmax=b_max, Nstep=Nstep, Nmeth=Nmeth,
                                      By=By, Bx=Bx, Energy=energy, **kwargs)

        for i, b in enumerate(self.By.T):
            dk = abs(b[3] ** 2 - b[4] ** 2 - b[2] ** 2) / abs(b[4])
            if dk > 1e-6:
                raise ValueError("Wiggler(H): kx^2 + kz^2 -ky^2 !=0, i = "
                                 "{0}".format(i))

        for i, b in enumerate(self.Bx.T):
            dk = abs(b[2] ** 2 - b[4] ** 2 - b[3] ** 2) / abs(b[4])
            if dk > 1e-6:
                raise ValueError("Wiggler(V): ky^2 + kz^2 -kx^2 !=0, i = "
                                 "{0}".format(i))

        self.NHharm = self.By.shape[1]
        self.NVharm = self.Bx.shape[1]


class QuantumDiffusion(_DictLongtMotion, Element):
    _BUILD_ATTRIBUTES = Element._BUILD_ATTRIBUTES + ['Lmatp']
    default_pass = {False: 'IdentityPass', True: 'QuantDiffPass'}
    _conversions = dict(Element._conversions, Lmatp=_array66)

    def __init__(self, family_name: str, lmatp: numpy.ndarray, **kwargs):
        """Quantum diffusion element

        Args:
            family_name:    Name of the element
            lmatp      :    Diffusion matrix for generation (see
                               ``at.physics.radiation.gen_quandiff_elem``)

        Default PassMethod: ``QuantDiffPass``
        """
        kwargs.setdefault('PassMethod', self.default_pass[True])
        super().__init__(family_name, Lmatp=lmatp, **kwargs)


def build_class_map():  # Missing class aliases (Bend)
    global CLASS_MAP

    def subclasses_recursive(cl):
        direct = cl.__subclasses__()
        indirect = []
        for subclass in direct:
            indirect.extend(subclasses_recursive(subclass))
        return frozenset([cl] + direct + indirect)

    cls_list = subclasses_recursive(Element)
    CLASS_MAP = {cls.__name__: cls for cls in cls_list}


def get_class_map():
    return CLASS_MAP


# build_class_map()

CLASS_MAP = dict((k, v) for k, v in locals().items()
                 if isinstance(v, type) and issubclass(v, Element))



"""
Analytical wake functions
"""
import numpy
from at.constants import clight
from scipy.interpolate import interp1d


def convolve_wakefun(srange, w, sigs):
    """Convolution of a wake function with a pulse of rms
    length sigs, this is use to generate a wake potential
    that can be added to the output of EM code like GDFIDL"""
    sigt = sigs/clight
    min_step = numpy.diff(srange)[0]
    t_out = numpy.arange(srange[0], srange[-1], min_step)
    sdiff = t_out[-1]-t_out[0]
    npoints = len(t_out)
    nt = npoints+npoints-1
    func = interp1d(srange, w, bounds_error=False, fill_value=0)
    wout = func(t_out)
    wout = numpy.append(wout, numpy.zeros(nt-len(wout)))
    fftr = numpy.fft.fft(wout)
    f = numpy.fft.fftshift(numpy.linspace(-(npoints-1)/sdiff,
                           (npoints-1)/sdiff, nt))
    fftl = numpy.exp(-(f*2*numpy.pi*sigt)**2/2)
    wout = numpy.fft.ifft(fftr*fftl)
    wout = numpy.roll(wout, int(npoints/2))
    t_out = numpy.linspace(t_out[0], t_out[-1], nt)
    func = interp1d(t_out, wout, bounds_error=False, fill_value=0)
    wout = func(srange)
    return wout


def long_resonator_wf(srange, frequency, qfactor, rshunt, beta):
    """Define the wake function (longitudinal) of a resonator
    with the given parameters according to Alex Chao's resonator
    model (Eq. 2.82) and definitions of the resonator in HEADTAIL.
    """
    omega = 2 * numpy.pi * frequency
    alpha = omega / (2 * qfactor)
    omegabar = numpy.sqrt(numpy.abs(omega**2 - alpha**2))
    dt = -srange/(beta * clight)
    if qfactor > 0.5:
        wake = (-(numpy.sign(dt) - 1) * rshunt * alpha *
                numpy.exp(alpha * dt) * (numpy.cos(omegabar * dt) +
                alpha / omegabar * numpy.sin(omegabar*dt)))
    elif qfactor == 0.5:
        wake = (-(numpy.sign(dt) - 1) * rshunt * alpha *
                numpy.exp(alpha * dt) * (1. + alpha * dt))
    elif qfactor < 0.5:
        wake = (-(numpy.sign(dt) - 1) * rshunt * alpha *
                numpy.exp(alpha * dt) * (numpy.cosh(omegabar * dt) +
                alpha / omegabar * numpy.sinh(omegabar * dt)))
    return wake


def transverse_resonator_wf(srange, frequency, qfactor, rshunt,
                            yokoya_factor, beta):
    """Define the wake function (transverse) of a resonator
    with the given parameters according to Alex Chao's
    resonator model (Eq. 2.82) and definitions of the resonator
    in HEADTAIL.
    """

    omega = 2 * numpy.pi * frequency
    alpha = omega / (2 * qfactor)
    omegabar = numpy.sqrt(numpy.abs(omega**2 - alpha**2))
    dt = -srange/(beta * clight)
    if qfactor > 0.5:
        wake = (-(numpy.sign(dt) - 1) * yokoya_factor * rshunt * omega**2 /
                (qfactor * omegabar) *
                numpy.exp(alpha*dt) * numpy.sin(omegabar*dt))
    elif qfactor == 0.5:
        wake = (-(numpy.sign(dt) - 1) * yokoya_factor * rshunt * omega**2 /
                qfactor *
                numpy.exp(alpha * dt) * dt)
    else:
        wake = (-(numpy.sign(dt) - 1) * yokoya_factor * rshunt * omega**2 /
                (qfactor * omegabar) *
                numpy.exp(alpha*dt) * numpy.sinh(omegabar*dt))
    return wake


def transverse_reswall_wf(srange, yokoya_factor, length, rvac, conduct, beta):
    """Define the wake function (transverse) of a resistive wall with the given
    parameters according to Alex Chao's RW model (2.53) and definitions used in
    HEADTAIL
    """

    if numpy.amin(srange) <= 0:
        raise ValueError("""
                        Provided srange has either negative values or 0s
                        This is not allowed for the transverse resistive wall
                        wake function. Please correct.
                        """)

    z0 = 119.9169832 * numpy.pi
    dt = -srange/(beta * clight)
    wake = (yokoya_factor * (numpy.sign(dt) - 1) / 2. *
            beta * length / numpy.pi / rvac**3 *
            numpy.sqrt(-z0 * clight / conduct / numpy.pi / dt))

    return wake



"""
Load lattices from Matlab files.
"""
from __future__ import print_function
import sys
from os.path import abspath, basename, splitext
from warnings import warn
from typing import Optional, Generator, Sequence
import scipy.io
import numpy
from ..lattice import elements, AtWarning, params_filter, AtError
from ..lattice import Element, Lattice
from .allfiles import register_format
from .utils import element_from_dict, element_from_m, RingParam
from .utils import element_to_dict, element_to_m

__all__ = ['load_mat', 'save_mat', 'load_m', 'save_m',
           'load_var']

_param_to_lattice = {'Energy': 'energy', 'Periodicity': 'periodicity',
                     'FamName': 'name', 'Particle': '_particle',
                     'cell_harmnumber': '_harmcell',
                     'HarmNumber': 'harmonic_number'}
_param_ignore = {'PassMethod', 'Length', 'cavpts'}

# Python to Matlab attribute translation
_matattr_map = dict(((v, k) for k, v in _param_to_lattice.items()))


def matfile_generator(params: dict, mat_file: str)\
        -> Generator[Element, None, None]:
    """Run through Matlab cells and generate AT elements

    Parameters:
        params:         Lattice building parameters (see :py:class:`.Lattice`)
        mat_file:       File name

    The following keys in ``params`` are used:

    ============    ===================
    **mat_key**     name of the Matlab variable containing the lattice.
                    Default: Matlab variable name if there is only one,
                    otherwise 'RING'
    **check**        Skip the coherence tests
    **quiet**       Suppress the warning for non-standard classes
    ============    ===================

    Yields:
        elem (Element): new Elements
    """
    def mclean(data):
        if data.dtype.type is numpy.str_:
            # Convert strings in arrays back to strings.
            return str(data[0]) if data.size > 0 else ''
        elif data.size == 1:
            v = data[0, 0]
            if issubclass(v.dtype.type, numpy.void):
                # Object => Return a dict
                return {f: mclean(v[f]) for f in v.dtype.fields}
            else:
                # Return a scalar
                return v
        else:
            # Remove any surplus dimensions in arrays.
            return numpy.squeeze(data)

    m = scipy.io.loadmat(params.setdefault('mat_file', mat_file))
    matvars = [varname for varname in m if not varname.startswith('__')]
    default_key = matvars[0] if (len(matvars) == 1) else 'RING'
    key = params.setdefault('mat_key', default_key)
    if key not in m.keys():
        kok = [k for k in m.keys() if '__' not in k]
        raise AtError('Selected mat_key does not exist, '
                      'please select in: {}'.format(kok))
    check = params.pop('check', True)
    quiet = params.pop('quiet', False)
    cell_array = m[key].flat
    for index, mat_elem in enumerate(cell_array):
        elem = mat_elem[0, 0]
        kwargs = {f: mclean(elem[f]) for f in elem.dtype.fields}
        yield element_from_dict(kwargs, index=index, check=check, quiet=quiet)


def ringparam_filter(params: dict, elem_iterator, *args)\
        -> Generator[Element, None, None]:
    """Run through all elements, process and optionally removes
    RingParam elements

    Parameters:
        params:         Lattice building parameters (see :py:class:`.Lattice`)
        elem_iterator:  Iterator over the lattice Elements

    Yields:
        elem (Element): new Elements

    The following keys in ``params`` are used:

    ============    ===================
    **keep_all**    keep RingParam elem_iterator as Markers
    ============    ===================

    The following keys in ``params`` are set:

    * ``name``
    * ``energy``
    * ``periodicity``
    * ``_harmnumber`` or
    * ``harmonic_number``
    * ``_particle``
    * ``_radiation``
    """
    keep_all = params.pop('keep_all', False)
    ringparams = []
    radiate = False
    for elem in elem_iterator(params, *args):
        if (elem.PassMethod.endswith('RadPass') or
                elem.PassMethod.endswith('CavityPass')):
            radiate = True
        if isinstance(elem, RingParam):
            ringparams.append(elem)
            for k, v in elem.items():
                if k not in _param_ignore:
                    params.setdefault(_param_to_lattice.get(k, k.lower()), v)
            if keep_all:
                pars = vars(elem).copy()
                name = pars.pop('FamName')
                yield elements.Marker(name, **pars)
        else:
            yield elem
    params['_radiation'] = radiate

    if len(ringparams) > 1:
        warn(AtWarning('More than 1 RingParam element, the 1st one is used'))


def load_mat(filename: str, **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice`  from a Matlab mat-file

    Parameters:
        filename:           Name of a '.mat' file

    Keyword Args:
        mat_key (str):      Name of the Matlab variable containing
          the lattice. Default: Matlab variable name if there is only one,
          otherwise 'RING'
        check (bool):       Run the coherence tests. Default:
          :py:obj:`True`
        quiet (bool):       Suppress the warning for non-standard
          classes. Default: :py:obj:`False`
        keep_all (bool):    Keep RingParam elements as Markers.
          Default: :py:obj:`False`
        name (str):         Name of the lattice. Default: taken from
          the lattice, or ``''``
        energy (float):     Energy of the lattice [eV]. Default: taken
          from the lattice elements
        periodicity(int):   Number of periods. Default: taken from the
          elements, or 1
        *:                  All other keywords will be set as Lattice
          attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
   """
    if 'key' in kwargs:  # process the deprecated 'key' keyword
        kwargs.setdefault('mat_key', kwargs.pop('key'))
    return Lattice(ringparam_filter, matfile_generator, abspath(filename),
                   iterator=params_filter, **kwargs)


def mfile_generator(params: dict, m_file: str)\
        -> Generator[Element, None, None]:
    """Run through the lines of a Matlab m-file and generate AT elements

    Parameters:
        params:         Lattice building parameters (see :py:class:`.Lattice`)
        m_file:         File name

    Yields:
        elem (Element): new Elements
"""
    with open(params.setdefault('m_file', m_file), 'rt') as file:
        _ = next(file)  # Matlab function definition
        _ = next(file)  # Cell array opening
        for lineno, line in enumerate(file):
            if line.startswith('};'):
                break
            try:
                elem = element_from_m(line)
            except ValueError:
                warn(AtWarning('Invalid line {0} skipped.'.format(lineno)))
                continue
            except KeyError:
                warn(AtWarning('Line {0}: Unknown class.'.format(lineno)))
                continue
            else:
                yield elem


def load_m(filename: str, **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice`  from a Matlab m-file

    Parameters:
        filename:           Name of a '.m' file

    Keyword Args:
        keep_all (bool):    Keep RingParam elements as Markers.
          Default: :py:obj:`False`
        name (str):         Name of the lattice. Default: taken from
          the lattice, or ``''``
        energy (float):     Energy of the lattice [eV]. Default: taken
          from the lattice elements
        periodicity(int):   Number of periods. Default: taken from the
          elements, or 1
        *:                  All other keywords will be set as Lattice
          attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
    """
    return Lattice(ringparam_filter, mfile_generator, abspath(filename),
                   iterator=params_filter, **kwargs)


def load_var(matlat: Sequence[dict], **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice` from a Matlab cell array

    Parameters:
        matlat:             Matlab lattice

    Keyword Args:
        keep_all (bool):    Keep RingParam elements as Markers.
          Default: :py:obj:`False`
        name (str):         Name of the lattice. Default: taken from
          the lattice, or ``''``
        energy (float):     Energy of the lattice [eV]. Default: taken
          from the lattice elements
        periodicity(int):   Number of periods. Default: taken from the
          elements, or 1
        *:                  All other keywords will be set as Lattice
          attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object
    """
    # noinspection PyUnusedLocal
    def var_generator(params, latt):
        for elem in latt:
            yield element_from_dict(elem)

    return Lattice(ringparam_filter, var_generator, matlat,
                   iterator=params_filter, **kwargs)


def matlab_ring(ring) -> Generator[Element, None, None]:
    """Prepend a RingParam element to a lattice"""
    dct = dict((_matattr_map.get(k, k.title()), v)
               for k, v in ring.attrs.items())
    famname = dct.pop('FamName')
    energy = dct.pop('Energy')
    yield RingParam(famname, energy, **dct)
    for elem in ring:
        yield elem


def save_mat(ring: Lattice, filename: str,
             mat_key: str = 'RING') -> None:
    """Save a :py:class:`.Lattice` as a Matlab mat-file

    Parameters:
        ring:           Lattice description
        filename:       Name of the '.mat' file
        mat_key (str):  Name of the Matlab variable containing
          the lattice. Default: ``'RING'``

    See Also:
        :py:func:`.save_lattice` for a generic lattice-saving function.
    """
    lring = tuple((element_to_dict(elem),) for elem in matlab_ring(ring))
    scipy.io.savemat(filename, {mat_key: lring}, long_field_names=True)


def save_m(ring: Lattice, filename: Optional[str] = None) -> None:
    """Save a :py:class:`.Lattice` as a Matlab m-file

    Parameters:
        ring:           Lattice description
        filename:       Name of the '.m' file. Default: outputs on
          :py:obj:`sys.stdout`

    See Also:
        :py:func:`.save_lattice` for a generic lattice-saving function.
    """

    def save(file):
        print('ring = {...', file=file)
        for elem in matlab_ring(ring):
            print(element_to_m(elem), file=file)
        print('};', file=file)

    if filename is None:
        save(sys.stdout)
    else:
        with open(filename, 'wt') as mfile:
            [funcname, _] = splitext(basename(filename))
            print('function ring = {0}()'.format(funcname), file=mfile)
            save(mfile)
            print('end', file=mfile)


register_format('.mat', load_mat, save_mat, descr='Matlab binary mat-file')
register_format('.m', load_m, save_m, descr='Matlab text m-file')



import numpy
from warnings import warn
# noinspection PyUnresolvedReferences
from .atpass import atpass as _atpass, elempass as _elempass
from ..lattice import DConstant, uint32_refpts


__all__ = ['lattice_pass', 'element_pass', 'atpass', 'elempass']

DIMENSION_ERROR = 'Input to lattice_pass() must be a 6xN array.'


# noinspection PyIncorrectDocstring
def lattice_pass(lattice, r_in, nturns=1, refpts=None, keep_lattice=False,
                 omp_num_threads=None, **kwargs):
    """
    lattice_pass(lattice, r_in, nturns=1, refpts=None, keep_lattice=False, keep_counter=False, turn=0, losses=False, omp_num_threads=None)

    lattice_pass tracks particles through each element of a lattice
    calling the element-specific tracking function specified in the Element's
    ``PassMethod`` field.

    Parameters:
        lattice (Iterable[Element]): list of elements
        r_in:                   6 x n_particles Fortran-ordered numpy array.
          On return, rin contains the final coordinates of the particles
        nturns (int):           number of turns to be tracked
        refpts (Uint32_refs):   numpy array of indices of elements where
          output is desired:

          * 0 means entrance of the first element
          * len(line) means end of the last element

    Keyword arguments:
        keep_lattice (Optional[bool]):  use elements persisted from a previous
          call. If True, assume that the lattice has not changed since
          the previous call.
        keep_counter (Optional[bool]):  Keep the turn number from the previous
          call.
        turn (Optional[int]):           Starting turn number. Ignored if
          keep_counter is True. The turn number is necessary to compute the
          absolute path length used in RFCavityPass.
        losses (Optional[bool]):        Boolean to activate loss maps output
        omp_num_threads (Optional[int]): number of OpenMP threads
          (default: automatic)

    The following keyword arguments overload the Lattice values

    Keyword arguments:
        particle (Optional[Particle]):  circulating particle.
          Default: ``lattice.particle`` if existing,
          otherwise ``Particle('relativistic')``
        energy (Optiona[float]):        lattice energy. Default 0.

    If ``energy`` is not available, relativistic tracking if forced,
    ``rest_energy`` is ignored.

    Returns:
        r_out: (6, N, R, T) array containing output coordinates of N particles
          at R reference points for T turns.

          If losses is True: {islost,turn,elem,coord} dictionary containing
          flag for particles lost (True -> particle lost), turn, element and
          coordinates at which the particle is lost. Set to zero for particles
          that survived

    .. note::

       * ``lattice_pass(lattice, r_in, refpts=len(line))`` is the same as
         ``lattice_pass(lattice, r_in)`` since the reference point len(line) is
         the exit of the last element.
       * ``lattice_pass(lattice, r_in, refpts=0)`` is a copy of ``r_in`` since
         the reference point 0 is the entrance of the first element.
       * To resume an interrupted tracking (for instance to get intermediate
         results), one must use one of the ``turn`` or ``keep_counter``
         keywords to ensure the continuity of the turn number.
       * For multiparticle tracking with large number of turn the size of
         ``r_out`` may increase excessively. To avoid memory issues
         ``lattice_pass(lattice, r_in, refpts=[])`` can be used. An empty list
         is returned and the tracking results of the last turn are stored in
         ``r_in``.

    """
    assert r_in.shape[0] == 6 and r_in.ndim in (1, 2), DIMENSION_ERROR
    if not isinstance(lattice, list):
        lattice = list(lattice)
    if refpts is None:
        refpts = len(lattice)
    if omp_num_threads is None:
        omp_num_threads = DConstant.omp_num_threads
    refs = uint32_refpts(refpts, len(lattice))
    bunch_currents = getattr(lattice, 'bunch_currents', numpy.zeros(1))
    bunch_spos = getattr(lattice, 'bunch_spos', numpy.zeros(1))
    kwargs.update({'bunch_currents': bunch_currents,
                   'bunch_spos': bunch_spos})
    # atpass returns 6xAxBxC array where n = x*y*z;
    # * A is number of particles;
    # * B is number of refpts
    # * C is the number of turns
    if r_in.flags.f_contiguous:
        return _atpass(lattice, r_in, nturns, refpts=refs,
                       reuse=keep_lattice,
                       omp_num_threads=omp_num_threads,
                       **kwargs)
    else:
        r_fin = numpy.asfortranarray(r_in)
        r_out = _atpass(lattice, r_fin, nturns, refpts=refs,
                        reuse=keep_lattice,
                        omp_num_threads=omp_num_threads,
                        **kwargs)
        r_in[:] = r_fin[:]
        return r_out


def element_pass(element, r_in, **kwargs):
    """Tracks particles through a single element.

    Parameters:
        element (Element):  AT element
        r_in:               (6, N) array: input coordinates of N particles.
          r_in is modified in-place and reports the coordinates at
          the end of the element. For the best efficiency, r_in
          should be given as F_CONTIGUOUS numpy array.

    Keyword arguments:
        particle (Optional[Particle]):  circulating particle.
          Default: Particle('relativistic')
        energy (Optiona[float]):        lattice energy

    If ``energy`` is not available, relativistic tracking if forced,
    ``rest_energy`` is ignored.

    Returns:
        r_out:              (6, N) array containing output the coordinates of
          the particles at the exit of the element.
    """
    r_in = numpy.asfortranarray(r_in)
    return _elempass(element, r_in, **kwargs)


# noinspection PyIncorrectDocstring
def atpass(*args, **kwargs):
    """
    atpass(line, r in, nturns, refpts=[], reuse=False, omp_num_threads=0)

    Track input particles ``rin`` along line for ``nturns`` turns.
    Record 6D phase space at elements corresponding to refpts for each turn.

    Parameters:
        line (Sequence[Element]): list of elements
        r_in:                   6 x n_particles Fortran-ordered numpy array.
          On return, rin contains the final coordinates of the particles
        nturns (int):           number of turns to be tracked

    Keyword arguments:
        refpts (Uint32_refs):   numpy array of indices of elements where
          output is desired:

          * 0 means entrance of the first element
          * len(line) means end of the last element

        energy:                 nominal energy [eV]
        rest_energy:            rest_energy of the particle [eV]
        charge:                 particle charge [elementary charge]
        reuse (Optional[bool]): if True, use previously cached description
          of the lattice.
        omp_num_threads (Optional[int]): number of OpenMP threads
          (default 0: automatic)
        losses (Optional[bool]):if True, process losses

    Returns:
        r_out:  6 x n_particles x n_refpts x n_turns Fortran-ordered
          numpy array of particle coordinates

    :meta private:
    """
    warn(UserWarning("The public interface for tracking is 'lattice_pass'"))
    return _atpass(*args, **kwargs)


# noinspection PyIncorrectDocstring
def elempass(*args, **kwargs):
    """elempass(element, rin)

    Tracks input particles rin through a single element.

    Parameters:
        element (Element):  AT element
        rin:                6 x n_particles Fortran-ordered numpy array.
          On return, rin contains the final coordinates of the particles

    Keyword arguments:
        energy:             nominal energy [eV]
        rest_energy:        rest_energy of the particle [eV]
        charge:             particle charge [elementary charge]

    :meta private:
    """
    warn(UserWarning("The public interface for tracking is 'element_pass'"))
    return _elempass(*args, **kwargs)



"""
Simple parallelisation of atpass() using multiprocessing.
"""
from functools import partial
import multiprocessing
from at.lattice import uint32_refpts
from warnings import warn
# noinspection PyUnresolvedReferences
from .atpass import atpass as _atpass
from at.lattice import AtWarning, DConstant
import numpy


__all__ = ['patpass']


globring = None


def format_results(results, r_in, losses):
    rin = [r['rin'] for r in results]
    r_in[:] = numpy.vstack(rin).T[:]
    if losses:
        rout = [r['results'][0] for r in results]
        rout = numpy.concatenate(rout, axis=1)
        lin = [r['results'][1] for r in results]
        lout = {}
        for k in lin[0].keys():
            lout[k] = numpy.hstack([li[k] for li in lin])
        return rout, lout
    else:
        rout = [r['results'] for r in results]
        rout = numpy.concatenate(rout, axis=1)
        return rout


def _atpass_one(ring, rin, **kwargs):
    if ring is None:
        result = _atpass(globring, rin, **kwargs)
    else:
        result = _atpass(ring, rin, **kwargs)
    return {'rin': rin, 'results': result}


def _pass(ring, r_in, pool_size, start_method, **kwargs):
    ctx = multiprocessing.get_context(start_method)
    if ctx.get_start_method() == 'fork':
        global globring
        globring = ring
        args = [(None, r_in[:, i]) for i in range(r_in.shape[1])]
        with ctx.Pool(pool_size) as pool:
            results = pool.starmap(partial(_atpass_one, **kwargs), args)
        globring = None
    else:
        args = [(ring, r_in[:, i]) for i in range(r_in.shape[1])]
        with ctx.Pool(pool_size) as pool:
            results = pool.starmap(partial(_atpass_one, **kwargs), args)
    losses = kwargs.pop('losses', False)
    return format_results(results, r_in, losses)


# noinspection PyIncorrectDocstring
def patpass(ring, r_in, nturns=1, refpts=None, pool_size=None,
            start_method=None, **kwargs):
    """
    patpass(lattice, r_in, nturns=1, refpts=None, keep_lattice=False,
    keep_counter=False, turn=0, losses=False, omp_num_threads=None,
    pool_size=None, start_method=None)

    Simple parallel implementation of atpass().  If more than one particle
    is supplied, use multiprocessing to run each particle in a separate
    process. In case a single particle is provided or the ring contains
    ImpedanceTablePass element, atpass() is returned

    ``patpass`` tracks particles through each element of a lattice
    calling the element-specific tracking function specified in the Element's
    ``PassMethod`` field.

    Parameters:
        lattice (Iterable[Element]): list of elements
        r_in:                   6 x n_particles Fortran-ordered numpy array.
          On return, rin contains the final coordinates of the particles
        nturns (int):           number of turns to be tracked
        refpts (Uint32_refs):   numpy array of indices of elements where
          output is desired:

          * 0 means entrance of the first element
          * len(line) means end of the last element

    Keyword arguments:
        keep_lattice (Optional[bool]):  use elements persisted from a previous
          call. If True, assume that the lattice has not changed since
          the previous call.
        keep_counter (Optional[bool]):  Keep the turn number from the previous
          call.
        turn (Optional[int]):           Starting turn number. Ignored if
          keep_counter is True. The turn number is necessary to compute the
          absolute path length used in RFCavityPass.
        losses (Optional[bool]):        Boolean to activate loss maps output
        pool_size (Optional[int]):      number of processes. If None,
          ``min(npart,nproc)`` is used
        start_method (Optional[str]):   This parameter allows to change the
          python multiprocessing start method, default=None uses the python
          defaults that is considered safe. Available parameters:
          '``fork'``, ``'spawn'``, ``'forkserver'``. Default for linux is
          ``'fork'``, default for macOS and  Windows is ``'spawn'``. ``'fork'``
          may used for macOS to speed up the calculation or to solve
          Runtime Errors, however it is considered unsafe.
        omp_num_threads (Optional[int]): number of OpenMP threads
          (default: automatic)

    The following keyword arguments overload the Lattice values

    Keyword arguments:
        particle (Optional[Particle]):  circulating particle.
          Default: ``lattice.particle`` if existing,
          otherwise ``Particle('relativistic')``
        energy (Optiona[float]):        lattice energy. Default 0.

    If ``energy`` is not available, relativistic tracking if forced,
    ``rest_energy`` is ignored.

    Returns:
        r_out: (6, N, R, T) array containing output coordinates of N particles
          at R reference points for T turns.

          If losses ==True: {islost,turn,elem,coord} dictionary containing
          flag for particles lost (True -> particle lost), turn, element and
          coordinates at which the particle is lost. Set to zero for particles
          that survived

    .. note::

       * For multiparticle tracking with large number of turn the size of
         ``r_out`` may increase excessively. To avoid memory issues
         ``patpass(lattice, r_in, refpts=[])`` can be used. An empty list
         is returned and the tracking results of the last turn are stored in
         ``r_in``.
       * By default, ``patpass`` will use all the available CPUs, to change
         the number of cores used in ALL functions using ``patpass``
         (``acceptance`` module for example) it is possible to set
         ``at.DConstant.patpass_poolsize`` to the desired value

    """
    def collective(rg) -> bool:
        """True if any element involves collective effects"""
        for elem in rg:
            if elem.is_collective:
                return True
        return False

    if not isinstance(ring, list):
        ring = list(ring)
    if refpts is None:
        refpts = len(ring)
    refpts = uint32_refpts(refpts, len(ring))
    bunch_currents = getattr(ring, 'bunch_currents', numpy.zeros(1))
    bunch_spos = getattr(ring, 'bunch_spos', numpy.zeros(1))
    kwargs.update(bunch_currents=bunch_currents, bunch_spos=bunch_spos)
    any_collective = collective(ring)
    if len(numpy.atleast_1d(r_in[0])) > 1 and not any_collective:
        if pool_size is None:
            pool_size = min(len(r_in[0]), multiprocessing.cpu_count(),
                            DConstant.patpass_poolsize)
        return _pass(ring, r_in, pool_size, start_method, nturns=nturns,
                     refpts=refpts, **kwargs)
    else:
        if any_collective:
            warn(AtWarning('Collective PassMethod found: use single process'))
        if r_in.flags.f_contiguous:
            return _atpass(ring, r_in, nturns=nturns,
                           refpts=refpts, **kwargs)
        else:
            r_fin = numpy.asfortranarray(r_in)
            r_out = _atpass(ring, r_fin, nturns=nturns,
                            refpts=refpts, **kwargs)
            r_in[:] = r_fin[:]
            return r_out



"""
Helper functions for working with AT lattices.

A lattice as understood by pyAT is any sequence of elements.  These functions
are useful for working with these sequences.

The ``refpts`` argument allow functions to select points in the lattice,
returned values are given at the entrance of each element specified in refpts;
``refpts`` can be:

#. an integer in the range [-len(ring), len(ring)-1] selecting the element
   according to python indexing rules. As a special case, len(ring) is
   allowed and refers to the end of the last element,
#. an ordered sequence of such integers without duplicates,
#. a sequence of booleans of maximum length len(ring)+1, where selected
   elements are True.
"""
import numpy
import functools
from typing import Callable, Optional, Sequence, Iterator, TypeVar
from typing import Union, Tuple, List
from itertools import compress
from fnmatch import fnmatch
from .elements import Element, Dipole

Refpts = TypeVar("Refpts", int, Sequence[int], bool, Sequence[bool])
ElementFilter = Callable[[Element], bool]
BoolRefpts = numpy.ndarray
Uint32Refpts = numpy.ndarray
Key = Union[type, Element, str]


__all__ = ['AtError', 'AtWarning', 'check_radiation', 'check_6d',
           'set_radiation', 'set_6d',
           'make_copy', 'uint32_refpts', 'bool_refpts',
           'checkattr', 'checktype', 'checkname',
           'get_cells', 'get_elements', 'get_refpts', 'get_s_pos',
           'refpts_count', 'refpts_len', 'refpts_iterator',
           'set_shift', 'set_tilt', 'tilt_elem', 'shift_elem',
           'get_value_refpts', 'set_value_refpts', 'Refpts',
           'get_geometry']


class AtError(Exception):
    pass


class AtWarning(UserWarning):
    pass


def check_radiation(rad: bool) -> Callable:
    """Deprecated decorator for optics functions (see :py:func:`check_6d`).

    :meta private:
    """
    return check_6d(rad)


def set_radiation(rad: bool) -> Callable:
    """Deprecated decorator for optics functions (see :py:func:`set_6d`).

    :meta private:
    """
    return set_6d(rad)


def check_6d(is_6d: bool) -> Callable:
    """Decorator for optics functions

    Wraps a function like ``func(ring, *args, **kwargs)`` where ring is a
    :py:class:`.Lattice` object. Raise :py:class:`.AtError` if ``ring.is_6d``
    is not the desired ``is_6d`` state. No test is performed if ring is not a
    :py:class:`.Lattice`.

    Parameters:
        is_6d: Desired 6D state

    Raises:
        AtError: if ``ring.is_6d`` is not ``is_6d``

    Example:

        .. code-block:: python

            @check_6d(False)
            def find_orbit4(ring, dct=None, guess=None, **kwargs):
                ...

        Raises :py:class:`.AtError` if ``ring.is_6d`` is :py:obj:`True`

    See Also:
        :py:func:`set_6d`
    """
    def radiation_decorator(func):
        @functools.wraps(func)
        def wrapper(ring, *args, **kwargs):
            ringrad = getattr(ring, 'is_6d', is_6d)
            if ringrad != is_6d:
                raise AtError('{0} needs radiation {1}'.format(
                    func.__name__, 'ON' if is_6d else 'OFF'))
            return func(ring, *args, **kwargs)
        return wrapper
    return radiation_decorator


def set_6d(is_6d: bool) -> Callable:
    """Decorator for optics functions

    Wraps a function like ``func(ring, *args, **kwargs)`` where ``ring`` is a
    :py:class:`.Lattice` object. If ``ring.is_6d`` is not the desired ``is_6d``
    state, ``func`` will be called with a modified copy of ``ring`` satisfying
    the ``is_6d`` state.

    Parameters:
        is_6d: Desired 6D state

    Example:

        .. code-block:: python

            @set_6d(True)
            def compute(ring)
                ...

        is roughly equivalent to:

        .. code-block:: python

            if ring.is_6d:
                compute(ring)
            else:
                compute(ring.enable_6d(copy=True))

    See Also:
        :py:func:`check_6d`, :py:meth:`.Lattice.enable_6d`,
        :py:meth:`.Lattice.disable_6d`
    """
    if is_6d:
        def setrad_decorator(func):
            @functools.wraps(func)
            def wrapper(ring, *args, **kwargs):
                rg = ring if ring.is_6d else ring.enable_6d(copy=True)
                return func(rg, *args, **kwargs)
            return wrapper
    else:
        def setrad_decorator(func):
            @functools.wraps(func)
            def wrapper(ring, *args, **kwargs):
                rg = ring.disable_6d(copy=True) if ring.is_6d else ring
                return func(rg, *args, **kwargs)
            return wrapper
    return setrad_decorator


def make_copy(copy: bool) -> Callable:
    """Decorator for optics functions

    Wraps a function like ``func(ring, refpts, *args, **kwargs)`` where
    ``ring`` is a :py:class:`.Lattice` object.

    If ``copy`` is :py:obj:`False`, the function is not modified,

    If ``copy`` is :py:obj:`True`, a shallow copy of ``ring`` is done, then the
    elements selected by ``refpts`` are deep-copied, then ``func`` is applied
    to the copy, and the new ring is returned.

    Parameters:
        copy:   If :py:obj:`True`, apply the decorated function to a copy of
          ``ring``
    """
    if copy:
        def copy_decorator(func):
            @functools.wraps(func)
            def wrapper(ring, refpts, *args, **kwargs):
                ring = ring.replace(refpts)
                func(ring, refpts, *args, **kwargs)
                return ring
            return wrapper
    else:
        def copy_decorator(func):
            return func
    return copy_decorator


def uint32_refpts(refpts: Refpts, n_elements: int) -> Uint32Refpts:
    """Returns an integer array of element indices, selecting ring elements.

    Parameters:
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element)
          #. a sequence of booleans marking the selected elements
        n_elements: Length of the lattice

    Returns:
        uint32_ref (Uint32Refpts): uint32 numpy array used for indexing
          ``Elements`` in a lattice.
    """
    refs = numpy.ravel(refpts)
    if (refpts is None) or (refs.size == 0):
        return numpy.array([], dtype=numpy.uint32)
    elif numpy.issubdtype(refs.dtype, numpy.bool_):
        return numpy.flatnonzero(refs).astype(numpy.uint32)

    # Handle negative indices
    refs = numpy.array([i if (i == n_elements) else i % n_elements
                        for i in refs], dtype=numpy.uint32)

    # Check ascending
    if refs.size > 1:
        prev = refs[0]
        for nxt in refs[1:]:
            if nxt < prev:
                raise ValueError('refpts should be given in ascending order')
            elif nxt == prev:
                raise ValueError('refpts contains duplicates or index(es) out'
                                 ' of range')
            prev = nxt

    return refs


# Private function accepting a callable for refpts
def _uint32_refs(ring: Sequence[Element], refpts: Refpts) -> Uint32Refpts:
    """Returns an integer array of element indices, selecting ring elements.

    Parameters:
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element)
          #. a sequence of booleans marking the selected elements
          #. a callable ``filtfunc`` such that ``filtfunc(elem)`` is True for
             selected elements

    Returns:
        uint32_ref (Uint32Refpts): uint32 numpy array used for indexing
          ``Elements`` in a lattice.
    """
    if callable(refpts):
        return numpy.array([i for i, el in enumerate(ring) if refpts(el)],
                           dtype=numpy.uint32)
    elif refpts is None:
        return numpy.array([], dtype=numpy.uint32)
    else:
        return uint32_refpts(refpts, len(ring))


def bool_refpts(refpts: Refpts, n_elements: int) -> BoolRefpts:
    """Returns n bool array of element indices, selecting ring elements.

    Parameters:
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element)
          #. a sequence of booleans marking the selected elements
        n_elements: Length of the lattice

    Returns:
        bool_refs (BoolRefpts):  A bool numpy array used for indexing
          ``Elements`` in a lattice.
    """
    refs = numpy.ravel(refpts)
    if (refpts is None) or (refs.size == 0):
        return numpy.zeros(n_elements + 1, dtype=bool)
    elif refs.dtype == bool:
        diff = 1 + n_elements - refs.size
        if diff <= 0:
            return refs[:n_elements + 1]
        else:
            return numpy.append(refs, numpy.zeros(diff, dtype=bool))
    else:
        brefpts = numpy.zeros(n_elements + 1, dtype=bool)
        brefpts[refs] = True
        return brefpts


# Private function accepting a callable for refpts
def _bool_refs(ring: Sequence[Element], refpts: Refpts) -> BoolRefpts:
    """Returns a bool array of element indices, selecting ring elements.

    Parameters:
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element)
          #. a sequence of booleans marking the selected elements
          #. a callable ``filtfunc`` such that ``filtfunc(elem)`` is True for
             selected elements

    Returns:
        bool_refs (BoolRefpts):  A bool numpy array used for indexing
          ``Elements`` in a lattice.
    """
    if callable(refpts):
        return numpy.array([refpts(el) for el in ring] + [False], dtype=bool)
    elif refpts is None:
        return numpy.zeros(len(ring) + 1, dtype=bool)
    else:
        return bool_refpts(refpts, len(ring))


def checkattr(attrname: str, attrvalue: Optional = None) \
        -> ElementFilter:
    # noinspection PyUnresolvedReferences
    """Checks the presence or the value of an attribute

        Returns a function to be used as an ``Element`` filter, which checks the
        presence or the value of an attribute of the provided ``Element``.
        This function can be used to extract from a ring all elements
        having a given attribute.

        Parameters:
            attrname: Attribute name
            attrvalue: Attribute value. If absent, the returned function checks
              the presence of an ``attrname`` attribute. If present, the
              returned function checks if ``attrname`` == ``attrvalue``.

        Returns:
            checkfun (ElementFilter):   Element filter function

        Examples:

            >>> cavs = filter(checkattr('Frequency'), ring)

            Returns an iterator over all elements in ring that have a
            ``Frequency`` attribute

            >>> elts = filter(checkattr('K', 0.0), ring)

            Returns an iterator over all elements in ring that have a ``K``
            attribute equal to 0.0
        """

    def testf(el):
        try:
            v = getattr(el, attrname)
            return (attrvalue is None) or (v == attrvalue)
        except AttributeError:
            return False

    return testf


def checktype(eltype: Union[type, Tuple[type, ...]]) -> ElementFilter:
    # noinspection PyUnresolvedReferences
    """Checks the type of an element

    Returns a function to be used as an ``Element`` filter, which checks the
    type of the provided ``Element``.
    This function can be used to extract from a ring all elements
    having a given type.

    Parameters:
        eltype: Desired ``Element`` type

    Returns:
        checkfun (ElementFilter):   Element filter function

    Examples:

        >>> qps = filter(checktype(at.Quadrupole), ring)

        Returns an iterator over all quadrupoles in ring
    """
    return lambda el: isinstance(el, eltype)


def checkname(pattern: str) -> ElementFilter:
    # noinspection PyUnresolvedReferences
    """Checks the name of an element

        Returns a function to be used as an ``Element`` filter, which checks the
        name of the provided ``Element``.
        This function can be used to extract from a ring all elements
        having a given name.

        Parameters:
            pattern: Desired ``Element`` name. Unix shell-style wildcards are
              supported (see ``fnmatch.fnmatch``)

        Returns:
            checkfun (ElementFilter):   Element filter function

        Examples:

            >>> qps = filter(checkname('QF*'), ring)

            Returns an iterator over all with name starting with ``QF``.
        """
    return lambda el: fnmatch(el.FamName, pattern)


# noinspection PyIncorrectDocstring
def get_cells(ring: Sequence[Element], *args) -> BoolRefpts:
    # noinspection PyUnresolvedReferences
    """
    get_cells(ring, filtfunc)
    get_cells(ring, attrname)
    get_cells(ring, attrname, attrvalue)
    Returns a bool array of element indices, selecting ring elements.

    Parameters:
        ring (Sequence[Element]):       Lattice description
        filtfunc (ElementFilter):   Filter function. Selects ``Elements``
          satisfying the filter function
        attrname (str):   Attribute name
        attrvalue (Any):  Attribute value. If absent, select the
          presence of an ``attrname`` attribute. If present, select
          ``Elements`` with ``attrname`` == ``attrvalue``.

    Returns:
        bool_refs (BoolRefpts):  numpy Array of ``bool`` with length
          len(ring)+1

    Examples:

        >>> refpts = get_cells(ring, 'Frequency')

        Returns a numpy array of booleans where all elements having a
        ``Frequency`` attribute are True

        >>> refpts = get_cells(ring, 'K', 0.0)

        Returns a numpy array of booleans where all elements having a ``K``
        attribute equal to 0.0 are True
    """
    testf = args[0] if callable(args[0]) else checkattr(*args)
    return numpy.append(numpy.fromiter(map(testf, ring), dtype=bool,
                                       count=len(ring)), False)


def refpts_iterator(ring: Sequence[Element], refpts: Refpts) \
        -> Iterator[Element]:
    """Return an iterator over selected elements in a lattice

    Parameters:
        ring:       Lattice description
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element)
          #. a sequence of booleans marking the selected elements
          #. a callable ``filtfunc`` such that ``filtfunc(elem)`` is True for
             selected elements

    Returns:
        elem_iter (Iterator[Element]):  Iterator over the elements in ``ring``
          selected by ``refpts``.
    """
    if refpts is None:
        return iter(())
    elif callable(refpts):
        return filter(refpts, ring)
    else:
        refs = numpy.ravel(refpts)
        if refs.size == 0:
            return iter(())
        elif refs.dtype == bool:
            return compress(ring, refs)
        else:
            return (ring[i] for i in refs)


# noinspection PyUnusedLocal
def refpts_count(refpts: Refpts, n_elements: int) -> int:
    """Returns the number of reference points

    Parameters:
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element)
          #. a sequence of booleans marking the selected elements
        n_elements: Lattice length

    Returns:
        nrefs (int):  The number of reference points
    """
    refs = numpy.ravel(refpts)
    if (refpts is None) or (refs.size == 0):
        return 0
    elif refs.dtype == bool:
        return numpy.count_nonzero(refs)
    else:
        return len(refs)


def refpts_len(ring: Sequence[Element], refpts: Refpts) -> int:
    """Returns the number of reference points

    Parameters:
        ring:       Lattice description
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element)
          #. a sequence of booleans marking the selected elements
          #. a callable ``filtfunc`` such that ``filtfunc(elem)`` is True for
             selected elements

    Returns:
        nrefs (int):  The number of reference points
    """
    if refpts is None:
        return 0
    elif callable(refpts):
        return len(list(filter(refpts, ring)))
    else:
        return refpts_count(refpts, len(ring))


def get_refpts(ring: Sequence[Element], key: Key, quiet=True) -> Uint32Refpts:
    """Returns an ``int`` array of element indices, selecting ring elements.

    Parameters:
        ring:   Lattice description
        key:    Element selection key. May be:

                  #. an element instance, will return all elements of the same
                     type in the lattice, e.g. key=Drift('d1', 1.0)
                  #. an element type, will return all elements of that type in
                     the lattice, e.g. ``at.elements.Sextupole``
                  #. a string to match against elements' ``FamName``, supports
                     Unix shell-style wildcards, e.g. ``'BPM_*1'``
        quiet:  if false print information about matched elements for FamName
               matches.

    Returns:
        uint32_refs (Uint32Refs):    numpy Array of ``int`` as long as the
          number of refpts

    See also:
        ``get_cells``
    """
    if isinstance(key, Element):
        checkfun = checktype(type(key))
    elif isinstance(key, type):
        checkfun = checktype(key)
    elif numpy.issubdtype(type(key), numpy.str_):
        checkfun = checkname(key)
        if not quiet:
            matched_fams = set(elem.FamName for elem in filter(checkfun, ring))
            ending = 'y' if len(matched_fams) == 1 else 'ies'
            print("String '{0}' matched {1} famil{2}: {3}\n"
                  "all corresponding elements have been "
                  "returned.".format(key, len(matched_fams), ending,
                                     ', '.join(matched_fams)))
    else:
        raise TypeError("Invalid key type {0}; please enter a string, element"
                        " type, or element instance.".format(type(key)))
    return uint32_refpts(list(map(checkfun, ring)), len(ring))


def get_elements(ring: Sequence[Element], key: Key, quiet=True) \
        -> list:
    """Returns a list of elements selected by ``key``.

    Parameters:
        ring:   Lattice description
        key:    Element selection key. May be:

                  #. an element instance, will return all elements of the same
                     type in the lattice, e.g. key=Drift('d1', 1.0)
                  #. an element type, will return all elements of that type in
                     the lattice, e.g. ``at.elements.Sextupole``
                  #. a string to match against elements' ``FamName``, supports
                     Unix shell-style wildcards, e.g. ``'BPM_*1'``
        quiet: if false print information about matched elements for FamName
               matches.

    Returns:
        elem_list (list):  list of ``Elements`` matching key
    """
    if isinstance(key, Element):
        checkfun = checktype(type(key))
    elif isinstance(key, type):
        checkfun = checktype(key)
    elif numpy.issubdtype(type(key), numpy.str_):
        checkfun = checkname(key)
        if not quiet:
            matched_fams = set(elem.FamName for elem in filter(checkfun, ring))
            ending = 'y' if len(matched_fams) == 1 else 'ies'
            print("String '{0}' matched {1} famil{2}: {3}\n"
                  "all corresponding elements have been "
                  "returned.".format(key, len(matched_fams), ending,
                                     ', '.join(matched_fams)))
    else:
        raise TypeError("Invalid key type {0}; please enter a string, element"
                        " type, or element instance.".format(type(key)))
    return list(filter(checkfun, ring))


def get_value_refpts(ring: Sequence[Element], refpts: Refpts,
                     attrname: str, index: Optional[int] = None):
    """Extracts attribute values from selected lattice ``Elements``.

    Parameters:
        ring:       Lattice description
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element)
          #. a sequence of booleans marking the selected elements
          #. a callable ``filtfunc`` such that ``filtfunc(elem)`` is True for
             selected elements
        attrname:   Attribute name
        index:      index of the value to retrieve if ``attrname`` is
          an array.

          If ``None`` the full array is retrieved

    Returns:
        attrvalues: numpy Array of attribute values.
    """
    if index is None:
        def getf(elem):
            return getattr(elem, attrname)
    else:
        def getf(elem):
            return getattr(elem, attrname)[index]

    return numpy.array([getf(elem) for elem in refpts_iterator(ring, refpts)])


def set_value_refpts(ring: Sequence[Element], refpts: Refpts,
                     attrname: str, attrvalues, index: Optional[int] = None,
                     increment: Optional[bool] = False,
                     copy: Optional[bool] = False):
    """Set the values of an attribute of an array of elements based on
    their refpts


    Parameters:
        ring:       Lattice description
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element)
          #. a sequence of booleans marking the selected elements
          #. a callable ``filtfunc`` such that ``filtfunc(elem)`` is True for
             selected elements
        attrname:   Attribute name
        attrvalues: Attribute values
        index:      index of the value to set if ``attrname`` is
          an array. if ``None``, the full array is replaced by ``attrvalue``
        increment:  Add values to the initial values.

          If False the initial value is replaced (Default)
        copy:       If ``False``, the modification is done in-place,

          If ``True``, return a shallow copy of the lattice. Only the
          modified elements are copied.

          .. Caution::

             a shallow copy means that all non-modified
             elements are shared with the original lattice.
             Any further modification will affect in both lattices.
    """
    if index is None:
        def setf(elem, value):
            setattr(elem, attrname, value)
    else:
        def setf(elem, value):
            getattr(elem, attrname)[index] = value

    if increment:
        attrvalues += get_value_refpts(ring, refpts, attrname, index=index)
    else:
        attrvalues = numpy.broadcast_to(attrvalues, (refpts_len(ring, refpts),))

    # noinspection PyShadowingNames
    @make_copy(copy)
    def apply(ring, refpts, values):
        for elm, val in zip(refpts_iterator(ring, refpts), values):
            setf(elm, val)

    return apply(ring, refpts, attrvalues)


def get_s_pos(ring: Sequence[Element], refpts: Optional[Refpts] = None) \
        -> Sequence[float]:
    """Returns the locations of selected elements

    Parameters:
        ring:       Lattice description
        refpts:     refpts may be:

          #. an integer or a sequence of integers
             (0 indicating the first element)
          #. a sequence of booleans marking the selected elements
          #. a callable ``filtfunc`` such that ``filtfunc(elem)`` is True for
             selected elements

    Returns:
        s_pos:  Array of locations of the elements selected by ``refpts``
    """
    if refpts is None:
        refpts = range(len(ring) + 1)
    # Positions at the end of each element.
    s_pos = numpy.cumsum([getattr(el, 'Length', 0.0) for el in ring])
    # Prepend position at the start of the first element.
    s_pos = numpy.concatenate(([0.0], s_pos))
    refpts = _uint32_refs(ring, refpts)
    return s_pos[refpts]


def tilt_elem(elem: Element, rots: float, relative: bool = False) -> None:
    r"""Set the tilt angle :math:`\theta` of an ``Element``

    The rotation matrices are stored in the ``R1`` and ``R2`` attributes

    :math:`R_1=\begin{pmatrix} cos\theta & sin\theta \\
    -sin\theta & cos\theta \end{pmatrix}`,
    :math:`R_2=\begin{pmatrix} cos\theta & -sin\theta \\
    sin\theta & cos\theta \end{pmatrix}`

    Parameters:
        elem:           Element to be tilted
        rots:           Tilt angle :math:`\theta` [rd].
          ``rots`` > 0 corresponds to a corkscrew rotation of the element
          looking in the direction of the beam
        relative:       If True, the rotation is added to the previous one
    """
    cs = numpy.cos(rots)
    sn = numpy.sin(rots)
    rm = numpy.asfortranarray(numpy.diag([cs, cs, cs, cs, 1.0, 1.0]))
    rm[0, 2] = sn
    rm[1, 3] = sn
    rm[2, 0] = -sn
    rm[3, 1] = -sn
    if relative and hasattr(elem, 'R1') and hasattr(elem, 'R2'):
        elem.R1 = elem.R1.dot(rm)
        elem.R2 = rm.T.dot(elem.R2)
    else:
        elem.R1 = rm
        elem.R2 = rm.T


def shift_elem(elem: Element, deltax: float = 0.0, deltaz: float = 0.0,
               relative: Optional[bool] = False) -> None:
    """Sets the transverse displacement of an ``Element``

    The translation vectors are stored in the ``T1`` and ``T2`` attributes

    Parameters:
        elem:           Element to be shifted
        deltax:         Horizontal displacement [m]
        deltaz:         Vertical displacement [m]
        relative:       If True, the translation is added to the previous one
    """
    tr = numpy.array([deltax, 0.0, deltaz, 0.0, 0.0, 0.0])
    if relative and hasattr(elem, 'T1') and hasattr(elem, 'T2'):
        elem.T1 -= tr
        elem.T2 += tr
    else:
        elem.T1 = -tr
        elem.T2 = tr


def set_tilt(ring: Sequence[Element], tilts, relative=False) -> None:
    """Sets the tilts of a list of elements.

    Parameters:
        ring:           Lattice description
        tilts:          Sequence of tilt values as long as ring or
          scalar tilt value applied to all elements
        relative:       If True, the rotation is added to the previous one
    """
    tilts = numpy.broadcast_to(tilts, (len(ring),))
    for el, tilt in zip(ring, tilts):
        tilt_elem(el, tilt, relative=relative)


def set_shift(ring: Sequence[Element], dxs, dzs, relative=False) -> None:
    """Sets the translations of a list of elements.

    Parameters:
        ring:           Lattice description
        dxs:            Sequence of horizontal displacements values as long as
          ring or scalar value applied to all elements [m]
        dzs:            Sequence of vertical displacements values as long as
          ring or scalar value applied to all elements [m]
        relative:       If True, the displacement is added to the previous one
    """
    dxs = numpy.broadcast_to(dxs, (len(ring),))
    dzs = numpy.broadcast_to(dzs, (len(ring),))
    for el, dx, dy in zip(ring, dxs, dzs):
        shift_elem(el, dx, dy, relative=relative)


def get_geometry(ring: List[Element],
                 start_coordinates: Tuple[float, float, float] = (0, 0, 0),
                 centered: bool = False):
    """Compute the 2D ring geometry in cartesian coordinates

    Parameters:
        ring: Lattice description
        start_coordinates: x,y,angle at starting point
        centered: it True the coordinates origin is the center of the ring

    Returns:
        geomdata: recarray containing, x, y, angle
        radius: machine radius

    Example:

       >>> geomdata, radius = get_geometry(ring)
    """

    geom_dtype = [('x', numpy.float64, (1, )),
                  ('y', numpy.float64, (1, )),
                  ('angle', numpy.float64, (1, ))]
    geomdata = numpy.recarray((len(ring)+1, ), dtype=geom_dtype)
    xx = numpy.zeros((len(ring)+1, 1))
    yy = numpy.zeros((len(ring)+1, 1))
    angle = numpy.zeros((len(ring)+1, 1))
    x, y, t = start_coordinates
    x0, y0, t0 = start_coordinates

    for ind, el in enumerate(ring+[ring[0]]):
        ll = el.Length
        if isinstance(el, Dipole) and el.BendingAngle != 0:
            ang = 0.5 * el.BendingAngle
            ll *= numpy.sin(ang)/ang
        else:
            ang = 0.0
        t -= ang
        x += ll * numpy.cos(t)
        y += ll * numpy.sin(t)
        t -= ang
        xx[ind] = x
        yy[ind] = y
        angle[ind] = t

    radius = get_s_pos(ring, len(ring))[0] / abs(t-t0) \
        if t != 0.0 else 0.0
    if centered:
        xx += -abs(radius)*numpy.sin(t0) - x0
        yy += abs(radius)*numpy.cos(t0) - y0
    geomdata['x'] = xx
    geomdata['y'] = yy
    geomdata['angle'] = angle
    return geomdata, radius



from at import elements, lattice_pass
import numpy
import pytest


@pytest.mark.parametrize('input_dim', [(0,), (5,), (7,), (1, 1), (6, 1, 1)])
def test_lattice_pass_raises_AssertionError_if_rin_incorrect_shape(input_dim):
    rin = numpy.zeros(input_dim)
    lattice = []
    with pytest.raises(AssertionError):
        lattice_pass(lattice, rin)


def test_multiple_particles_lattice_pass():
    lattice = [elements.Drift('Drift', 1.0)]
    rin = numpy.zeros((6, 2))
    rin[0, 0] = 1e-6  # particle one offset in x
    rin[2, 1] = 1e-6  # particle two offset in y
    r_original = numpy.copy(rin)
    r_out = lattice_pass(lattice, rin, nturns=2)
    # particle position is not changed passing through the drift
    numpy.testing.assert_equal(r_original[:, 0], r_out[:, 0, 0, 0])
    numpy.testing.assert_equal(r_original[:, 0], r_out[:, 0, 0, 1])
    numpy.testing.assert_equal(r_original[:, 1], r_out[:, 1, 0, 0])
    numpy.testing.assert_equal(r_original[:, 1], r_out[:, 1, 0, 1])


def test_lattice_convert_to_list_if_incorrect_type():
    lattice = numpy.array([elements.Drift('Drift', 1.0)])
    rin = numpy.zeros((6, 2))
    rin[0, 0] = 1e-6
    r_original = numpy.copy(rin)
    r_out = lattice_pass(lattice, rin, 1)
    numpy.testing.assert_equal(r_original, r_out.reshape(6, 2))



"""AT plotting functions"""
from typing import Tuple
from at.lattice import Lattice
import matplotlib.pyplot as plt
import numpy
from numpy import ndarray


# Function to compute and plot acceptance
def plot_acceptance(ring: Lattice, *args, **kwargs):
    # noinspection PyUnresolvedReferences
    r"""Plots the acceptance

    Computes the acceptance at repfts observation points using
    :py:func:`.get_acceptance` and plots the tracked
    and survived particles, and the acceptance boundary.

    Parameters:
        ring:           Lattice definition

    Keyword Args:
        acceptance(Tuple[ndarray, ndarray, ndarray]): Tuple containing
          pre-computed acceptance ``(boundary, survived, grid)``
        planes:         max. dimension 2, Plane(s) to scan for the acceptance.
          Allowed values are: ``'x'``, ``'xp'``, ``'y'``,
          ``'yp'``, ``'dp'``, ``'ct'``
        npoints:        (len(planes),) array: number of points in each
          dimension
        amplitudes:     (len(planes),) array: set the search range:

          * :py:attr:`GridMode.CARTESIAN/RADIAL <.GridMode.RADIAL>`:
            max. amplitude
          * :py:attr:`.GridMode.RECURSIVE`: initial step
        nturns (int):       Number of turns for the tracking
        obspt (Refpts):    Observation points. Default: start of the machine
        dp (float):         Static momentum offset
        offset:             Initial orbit. Default: closed orbit
        bounds:             Defines the tracked range: range=bounds*amplitude.
          It can be used to select quadrants. For example, default values are:

          * :py:attr:`.GridMode.CARTESIAN`: ((-1, 1), (0, 1))
          * :py:attr:`GridMode.RADIAL/RECURSIVE <.GridMode.RADIAL>`: ((0, 1),
            (:math:`\pi`, 0))
        grid_mode (GridMode):   Defines the evaluation grid:

          * :py:attr:`.GridMode.CARTESIAN`: full [:math:`\:x, y\:`] grid
          * :py:attr:`.GridMode.RADIAL`: full [:math:`\:r, \theta\:`] grid
          * :py:attr:`.GridMode.RECURSIVE`: radial recursive search
        use_mp (bool):      Use python multiprocessing (:py:func:`.patpass`,
          default use :py:func:`.lattice_pass`). In case multiprocessing is not
          enabled, ``grid_mode`` is forced to :py:attr:`.GridMode.RECURSIVE`
          (most efficient in single core)
        verbose (bool):     Print out some information
        divider (int):      Value of the divider used in
          :py:attr:`.GridMode.RECURSIVE` boundary search
        shift_zero:
        start_method (str): Python multiprocessing start method. The default
          :py:obj:`None` uses the python default that is considered safe.
          Available parameters: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          The default for linux is ``'fork'``, the default for macOS and
          Windows is ``'spawn'``. ``'fork'`` may used for macOS to speed up
          the calculation or to solve runtime errors, however  it is
          considered unsafe.

    Returns:
        boundary:   (2,n) array: 2D acceptance
        tracked:    (2,n) array: Coordinates of tracked particles
        survived:   (2,n) array: Coordinates of surviving particles

    Example:
        >>> ring.plot_acceptance(planes, npoints, amplitudes)
        >>> plt.show()
    """

    units = {'x': '[m]', 'xp': '[rad]', 'y': '[m]',
             'yp': '[rad]', 'dp': '', 'ct': '[m]'}

    obspt = kwargs.pop('obspt', None)
    block = kwargs.pop('block', False)
    acceptance = kwargs.pop('acceptance', None)
    if obspt is not None:
        assert numpy.isscalar(obspt), 'Scalar value needed for obspt'
    kwargs['refpts'] = obspt
    if acceptance is None:
        boundary, survived, grid = ring.get_acceptance(*args, **kwargs)
    else:
        boundary, survived, grid = acceptance
    planes = args[0]
    plt.figure()
    plt.plot(*grid, '.', label='Tracked particles')
    plt.plot(*survived, '.', label='Survived particles')
    if len(planes) == 1:
        plt.plot(boundary, numpy.zeros(2), label='Acceptance')
        plt.title('1D {0} acceptance'.format(planes[0]))
        plt.xlabel('{0} {1}'.format(planes[0], units[planes[0]]))
    else:
        plt.plot(*boundary, label='Acceptance')
        plt.title('2D {0} {1} acceptance'.format(planes[0], planes[1]))
        plt.xlabel('{0} {1}'.format(planes[0], units[planes[0]]))
        plt.ylabel('{0} {1}'.format(planes[1], units[planes[1]]))
    plt.legend()
    plt.show(block=block)
    return boundary, survived, grid


def plot_geometry(ring: Lattice, start_coordinates=(0, 0, 0),
                  centered=False, ax=None, label=''):
    """Compute the 2D ring geometry in cartesian coordinates.
    
    Usage:
    
    .. code-block:: python
    
       ring.plot_geometry()

    Parameters:
        ring: Lattice description
        start_coordinates: x,y,angle at starting point
        centered: it True the coordinates origin is the center of the ring
        offset: (dx, dy) offsets coordinates by the given amount
        ax: axes where to plot, if not given axes are created
        label: label of curve

    Returns:
        geomdata: recarray containing, x, y, angle
        radius: machine radius
        ax: plot axis
    """
    if not ax:
        fig, ax = plt.subplots()
    geom, radius = ring.get_geometry(start_coordinates=start_coordinates,
                                     centered=centered)
    ax.plot(geom['x'], geom['y'], 'o:', linewidth=0.5, markersize=2,
            label=label)
    ax.set_xlabel('x [m]')
    ax.set_ylabel('y [m]')
    ax.set_aspect('equal', 'box')
    return geom, radius, ax


Lattice.plot_acceptance = plot_acceptance
Lattice.plot_geometry = plot_geometry



from at import linopt6
import numpy
import matlab
from numpy.testing import assert_allclose as assert_close
import pytest


def _py_data(ml_data):
    """Convert a Matlab vector to a numpy vector"""
    return numpy.squeeze(numpy.asarray(ml_data))


def _ml_refs(ring, py_refs):
    """Convert refpoints to Matlab"""
    uintrefs = ring.uint32_refpts(py_refs)
    return matlab.double([float(ref+1) for ref in uintrefs])


def _compare_4d(py_data, ml_data, fields, atol=1.e-12, rtol=1.e-7):
    for (ml_key, py_key) in fields:
        ml_val = _py_data(ml_data[ml_key])
        if py_key == 'closed_orbit':
            py_val = numpy.squeeze(py_data[py_key][:, :4])
        else:
            py_val = numpy.squeeze(py_data[py_key])
        assert_close(py_val, ml_val, atol=atol, rtol=rtol)


def _compare_6d(py_data, ml_data, fields, atol=1.e-12, rtol=1.e-7):
    for (ml_key, py_key) in fields:
        ml_val = _py_data(ml_data[ml_key])
        py_val = numpy.squeeze(py_data[py_key])
        assert_close(py_val, ml_val, atol=atol, rtol=rtol)


@pytest.mark.parametrize('dp', (-0.01, 0.0, 0.01))
@pytest.mark.parametrize('lattices',
                         [pytest.lazy_fixture('hmba'),
                          pytest.lazy_fixture('dba')])
def test_4d_analysis(engine, lattices, dp):
    """Compare linopt6 in 4D"""
    py_lattice, ml_lattice, _ = lattices
    fields = [('SPos', 's_pos'), ('ClosedOrbit', 'closed_orbit'),
              ('M', 'M'), ('A', 'A'), ('Dispersion', 'dispersion'),
              ('alpha', 'alpha'), ('beta', 'beta'), ('mu', 'mu'),
              ]
    pypts = range(10)
    mlpts = _ml_refs(py_lattice, pypts)

    # python call
    py_data0, py_ringdata, py_data = linopt6(py_lattice, refpts=pypts,
                                             dp=dp, get_w=True)
    # Matlab call
    ml_ringdata, ml_data = engine.pyproxy('atlinopt6', ml_lattice, mlpts,
                                          'dp', dp, 'get_w', nargout=2)
    # Comparison
    assert_close(py_ringdata.tune,
                 _py_data(ml_ringdata['tune']),
                 atol=1.e-8, rtol=0)
    assert_close(py_ringdata.chromaticity,
                 _py_data(ml_ringdata['chromaticity']),
                 atol=5.e-4, rtol=0)

    _compare_6d(py_data, ml_data, fields, atol=1.e-6, rtol=1e-8)
    _compare_6d(py_data, ml_data, [('W', 'W')], atol=1.e-6, rtol=1e-4)


@pytest.mark.parametrize('lattices',
                         [pytest.lazy_fixture('hmba')])
def test_6d_analysis(engine, lattices):
    """Compare linopt6 in 6D"""
    py_lattice, ml_lattice, _ = lattices
    fields = [('SPos', 's_pos'), ('ClosedOrbit', 'closed_orbit'),
              ('M', 'M'), ('A', 'A'), ('Dispersion', 'dispersion'),
              ('alpha', 'alpha'), ('beta', 'beta'), ('mu', 'mu'),
              ('W', 'W')
              ]
    ml_lattice = engine.pyproxy('atradon', ml_lattice)
    py_lattice = py_lattice.radiation_on(copy=True)
    pypts = range(10)
    mlpts = _ml_refs(py_lattice, pypts)

    # python call
    py_data0, py_ringdata, py_data = linopt6(py_lattice, refpts=pypts,
                                             get_w=True)
    # Matlab call
    ml_ringdata, ml_data = engine.pyproxy('atlinopt6', ml_lattice, mlpts,
                                          'get_w', nargout=2)
    # Comparison
    assert_close(py_ringdata.tune,
                 _py_data(ml_ringdata['tune']),
                 atol=1.e-12, rtol=0)
    assert_close(py_ringdata.chromaticity,
                 _py_data(ml_ringdata['chromaticity']),
                 atol=1.e-5, rtol=0)

    _compare_6d(py_data, ml_data, fields, atol=1.e-6, rtol=1e-8)



"""Text representation of a python AT lattice with each element represented by
its :py:func:`repr` string
"""
from __future__ import print_function
import sys
from os.path import abspath
from typing import Optional
import numpy
from at.lattice import Lattice
from at.load import register_format
from at.load.utils import element_from_string
# imports necessary in' globals()' for 'eval'
# noinspection PyUnresolvedReferences
from at.lattice import Particle

__all__ = ['load_repr', 'save_repr']


def load_repr(filename: str, **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice`  from a text repr-file

    Parameters:
        filename:           Name of a '.m' file

    Keyword Args:
        name (str):         Name of the lattice. Default: taken from
          the lattice
        energy (float):     Energy of the lattice [eV]. Default: taken
          from the lattice elements
        periodicity(int):   Number of periods. Default: taken from the
          elements, or 1
        *:                  All other keywords will be set as Lattice
          attributes

    Returns:
        lattice (Lattice):          New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
    """
    def elem_iterator(params, repr_file):
        with open(params.setdefault('repr_file', repr_file), 'rt') as file:
            # the 1st line is the dictionary of saved lattice parameters
            for k, v in eval(next(file)).items():
                params.setdefault(k, v)
            for line in file:
                yield element_from_string(line.strip())

    return Lattice(abspath(filename), iterator=elem_iterator, **kwargs)


def save_repr(ring: Lattice, filename: Optional[str] = None) -> None:
    """Save a :py:class:`.Lattice` as a repr-file

    Parameters:
        ring:           Lattice description
        filename:       Name of the '.repr' file. Default: outputs on
          :py:obj:`sys.stdout`

    See Also:
        :py:func:`.save_lattice` for a generic lattice-saving function.
    """
    def save(file):
        # print(repr(dict((k, v) for k, v in vars(ring).items()
        #                 if not k.startswith('_'))), file=file)
        print(repr(ring.attrs), file=file)
        for elem in ring:
            print(repr(elem), file=file)

    # Save the current options
    opts = numpy.get_printoptions()
    # Set options to print the full representation of float variables
    numpy.set_printoptions(formatter={'float_kind': repr})
    if filename is None:
        save(sys.stdout)
    else:
        with open(filename, 'wt') as reprfile:
            save(reprfile)
    # Restore the current options
    numpy.set_printoptions(**opts)


register_format('.repr', load_repr, save_repr,
                descr='Text representation of a python AT Lattice')



import sys
import numpy
from io import BytesIO, StringIO
from at.lattice import elements, uint32_refpts, bool_refpts, checkattr
from at.lattice import checktype, get_cells, refpts_iterator, get_elements
from at.lattice import get_s_pos, tilt_elem, shift_elem, set_tilt, set_shift
import pytest


@pytest.mark.parametrize('ref_in, expected', (
    [2, numpy.array([2], dtype=numpy.uint32)],
    [-1, numpy.array([4], dtype=numpy.uint32)],
    [7, numpy.array([2], dtype=numpy.uint32)],
    [[1, 2, 3], numpy.array([1, 2, 3], dtype=numpy.uint32)],
    [[0, 1, 2, 3, 4, 5], numpy.array([0, 1, 2, 3, 4, 5], dtype=numpy.uint32)],
    [[0, 6, 2, -2, 4, 5], numpy.array([0, 1, 2, 3, 4, 5], dtype=numpy.uint32)],
))
def test_uint32_refpts_converts_numerical_inputs_correctly(ref_in, expected):
    numpy.testing.assert_equal(uint32_refpts(ref_in, 5), expected)
    ref_in2 = numpy.asarray(ref_in).astype(numpy.float64)
    numpy.testing.assert_equal(uint32_refpts(ref_in2, 5), expected)
    ref_in2 = numpy.asarray(ref_in).astype(numpy.int64)
    numpy.testing.assert_equal(uint32_refpts(ref_in2, 5), expected)


@pytest.mark.parametrize('ref_in, expected', (
    [None, numpy.array([], dtype=numpy.uint32)],
    [[], numpy.array([], dtype=numpy.uint32)],
    [True, numpy.array([0], dtype=numpy.uint32)],
    [False, numpy.array([], dtype=numpy.uint32)],
    [[True, False, True], numpy.array([0, 2], dtype=numpy.uint32)],
    [[False, False, False, False, False, False],
     numpy.array([], dtype=numpy.uint32)],
    [[True, True, True, True, True, True],
     numpy.array([0, 1, 2, 3, 4, 5], dtype=numpy.uint32)]
))
def test_uint32_refpts_converts_other_input_types_correctly(ref_in, expected):
    numpy.testing.assert_equal(uint32_refpts(ref_in, 5), expected)


# too long, misordered, duplicate, -ve indexing misordered, -ve indexing
# duplicate, wrap-around indexing misordered, wrap-around indexing duplicate,
# -ve indexing and wrap-around indexing duplicate, -ve indexing and wrap-around
# indexing misordered
@pytest.mark.parametrize('ref_in', ([0, 1, 2, 3], [2, 1], [0, 0], [-1, 0],
                                    [0, -2], [3, 0], [1, 3], [-1, 3], [3, -2]))
def test_uint32_refpts_throws_ValueError_correctly(ref_in):
    with pytest.raises(ValueError):
        uint32_refpts(ref_in, 2)


@pytest.mark.parametrize('ref_in, expected', (
    [[True], numpy.array([True, False, False, False], dtype=bool)],
    [[], numpy.array([False, False, False, False], dtype=bool)],
    [None, numpy.array([False, False, False, False], dtype=bool)],
    [2, numpy.array([False, False, True, False], dtype=bool)],
    [[1, 2], numpy.array([False, True, True, False], dtype=bool)],
))
def test_bool_refpts(ref_in, expected):
    numpy.testing.assert_equal(bool_refpts(ref_in, 3), expected)


@pytest.mark.parametrize('ref_in', (10, range(100)))
def test_bool_refpts_throws_IndexErrorrs(ref_in):
    with pytest.raises(IndexError):
        bool_refpts(ref_in, 3)


def test_checkattr(simple_ring):
    assert checkattr('Length')(simple_ring[0]) is True
    assert checkattr('not_an_attr')(simple_ring[0]) is False
    assert (list(filter(checkattr('Length', 1), simple_ring)) ==
            [simple_ring[0], simple_ring[3], simple_ring[4], simple_ring[5]])
    assert list(filter(checkattr('Length', 2), simple_ring)) == []
    assert list(filter(checkattr('not_an_attr'), simple_ring)) == []


def test_checktype(simple_ring):
    assert checktype(elements.Drift)(simple_ring[0]) is True
    assert checktype(elements.Marker)(simple_ring[0]) is False
    assert (list(filter(checktype(elements.Drift), simple_ring)) ==
            [simple_ring[0], simple_ring[3], simple_ring[4], simple_ring[5]])
    assert list(filter(checktype(elements.Monitor), simple_ring)) == []


def test_get_cells(simple_ring):
    a = numpy.array([True, True, True, True, True, True, False])
    numpy.testing.assert_equal(get_cells(simple_ring, checkattr('Length')), a)
    a = numpy.array([False, True, False, False, False, False, False])
    numpy.testing.assert_equal(get_cells(simple_ring, 'attr'), a)
    a = numpy.array([True, False, False, False, False, False, False])
    numpy.testing.assert_equal(get_cells(simple_ring, 'FamName', 'D1'), a)


def test_refpts_iterator(simple_ring):
    assert (list(refpts_iterator(simple_ring, [0, 1, 2, 3, 4, 5])) ==
            simple_ring)
    assert (list(refpts_iterator(simple_ring,
                                 numpy.ones(6, dtype=bool))) ==
            simple_ring)
    assert list(refpts_iterator(simple_ring, [1])) == [simple_ring[1]]
    a = numpy.array([False, True, False, False, False, False])
    assert list(refpts_iterator(simple_ring, a)) == [simple_ring[1]]


def test_get_elements(hmba_lattice):
    # test FamName direct match
    assert get_elements(hmba_lattice, 'BPM_06') == [hmba_lattice[65]]
    # test FamName wildcard matching
    assert get_elements(hmba_lattice, 'QD2?') == list(hmba_lattice[9, 113])
    assert get_elements(hmba_lattice, 'QD3*') == list(hmba_lattice[19, 105])
    assert get_elements(hmba_lattice, 'S*H2B') == list([hmba_lattice[55]])
    assert get_elements(hmba_lattice, '*C_1') == list(hmba_lattice[59, 60])
    assert get_elements(hmba_lattice,
                        'DR_2[1-3]') == list(hmba_lattice[54, 56, 58])
    assert get_elements(hmba_lattice,
                        'DR_2[!1-7]') == list(hmba_lattice[52, 78, 80])
    # test element instance
    marker = elements.Marker('M1')
    assert get_elements(hmba_lattice,
                        marker) == list(hmba_lattice[1, 12, 61, 67, 73])
    # test element type
    assert get_elements(hmba_lattice, elements.RFCavity) == [hmba_lattice[0]]
    # test invalid key raises TypeError
    with pytest.raises(TypeError):
        get_elements(hmba_lattice, None)
    # test quiet suppresses print statement correctly
    if sys.version_info < (3, 0):
        capturedOutput = BytesIO()
    else:
        capturedOutput = StringIO()
    sys.stdout = capturedOutput
    get_elements(hmba_lattice, 'BPM_06', quiet=True)
    sys.stdout = sys.__stdout__
    assert capturedOutput.getvalue() == ''
    sys.stdout = capturedOutput
    get_elements(hmba_lattice, 'BPM_06', quiet=False)
    sys.stdout = sys.__stdout__
    assert capturedOutput.getvalue() == ("String 'BPM_06' matched 1 family: "
                                         "BPM_06\nall corresponding elements "
                                         "have been returned.\n")


def test_get_s_pos_returns_zero_for_empty_lattice():
    numpy.testing.assert_equal(get_s_pos([], None), numpy.array((0,)))


def test_get_s_pos_returns_length_for_lattice_with_one_element():
    e = elements.LongElement('e', 0.1)
    assert get_s_pos([e], [1]) == numpy.array([0.1])


def test_get_s_pos_returns_all_pts_for_lat_with_2_elements_and_refpts_None():
    e = elements.LongElement('e', 0.1)
    f = elements.LongElement('e', 2.1)
    print(get_s_pos([e, f], None))
    numpy.testing.assert_equal(get_s_pos([e, f], None),
                               numpy.array([0, 0.1, 2.2]))


def test_get_s_pos_returns_all_pts_for_lat_with_2_elements_using_int_refpts():
    e = elements.LongElement('e', 0.1)
    f = elements.LongElement('e', 2.1)
    lat = [e, f]
    numpy.testing.assert_equal(get_s_pos(lat, range(len(lat) + 1)),
                               numpy.array([0, 0.1, 2.2]))


def test_get_s_pos_returns_all_pts_for_lat_with_2_elements_using_bool_refpts():
    e = elements.LongElement('e', 0.1)
    f = elements.LongElement('e', 2.1)
    lat = [e, f]
    numpy.testing.assert_equal(get_s_pos(lat, numpy.ones(3, dtype=bool)),
                               numpy.array([0, 0.1, 2.2]))


def test_tilt_elem(simple_ring):
    tilt_elem(simple_ring[0], (numpy.pi/4))
    v = 1/2**0.5
    a = numpy.diag([v, v, v, v, 1.0, 1.0])
    a[0, 2], a[1, 3], a[2, 0], a[3, 1] = v, v, -v, -v
    numpy.testing.assert_allclose(simple_ring[0].R1, a)
    numpy.testing.assert_allclose(simple_ring[0].R2, a.T)


def test_shift_elem(simple_ring):
    shift_elem(simple_ring[2], 1.0, 0.5)
    a = numpy.array([1.0, 0.0, 0.5, 0.0, 0.0, 0.0])
    numpy.testing.assert_equal(simple_ring[2].T1, -a)
    numpy.testing.assert_equal(simple_ring[2].T2, a)


def test_set_tilt(simple_ring):
    set_tilt(simple_ring, [(numpy.pi/4), 0, 0, 0, (numpy.pi/4), 0])
    v = 1/2**0.5
    a = numpy.diag([v, v, v, v, 1.0, 1.0])
    a[0, 2], a[1, 3], a[2, 0], a[3, 1] = v, v, -v, -v
    numpy.testing.assert_allclose(simple_ring[0].R1, a)
    numpy.testing.assert_allclose(simple_ring[0].R2, a.T)
    numpy.testing.assert_allclose(simple_ring[0].R1, a)
    numpy.testing.assert_allclose(simple_ring[0].R2, a.T)
    ring = [simple_ring[0]]
    set_tilt(ring, (0))
    numpy.testing.assert_allclose(ring[0].R1, numpy.eye(6))
    numpy.testing.assert_allclose(ring[0].R2, numpy.eye(6))


def test_set_shift(simple_ring):
    set_shift(simple_ring, numpy.array([0., 0., 0., 1., 0., 0.5]),
              numpy.array([0., 0., 0., 2., 0., 1.]))
    a = numpy.array([0.5, 0., 1., 0., 0., 0.])
    numpy.testing.assert_equal(simple_ring[3].T1, -a*2)
    numpy.testing.assert_equal(simple_ring[3].T2, a*2)
    numpy.testing.assert_equal(simple_ring[5].T1, -a)
    numpy.testing.assert_equal(simple_ring[5].T2, a)
    ring = [simple_ring[3]]
    set_shift(ring, 3, 5)
    a = numpy.array([3., 0., 5., 0., 0., 0.])
    numpy.testing.assert_equal(simple_ring[3].T1, -a)
    numpy.testing.assert_equal(simple_ring[3].T2, a)



from setuptools import build_meta as _orig
import os

# Mandatory hooks
build_sdist = _orig.build_sdist


def _get_option(config, key, default_value):
    val = os.environ.get(key.upper(), default_value)
    if config is not None:
        val = config.get(key.lower(), val)
    return eval(val)


def build_wheel(wheel_dir, config_settings=None, metadata_dir=None):
    print("** Entering build_wheel. "
          "Config settings:", config_settings)
    os.environ["MPI"] = str(_get_option(config_settings, 'MPI', '0'))
    os.environ["OPENMP"] = str(_get_option(config_settings, 'OPENMP', '0'))
    omp_threshold = _get_option(config_settings,
                                'OMP_PARTICLE_THRESHOLD', 'None')
    if omp_threshold is not None:
        os.environ["OMP_PARTICLE_THRESHOLD"] = str(omp_threshold)
    print("** MPI:", os.environ.get('MPI', 'None'))
    print("** OPENMP:", os.environ.get('OPENMP', 'None'))
    ret = _orig.build_wheel(wheel_dir, config_settings, metadata_dir)
    print("** Leaving build_wheel")
    return ret


# Optional hook
def get_requires_for_build_wheel(config_settings=None):
    print("** Entering get_requires_for_build_wheel. "
          "Config settings:", config_settings)
    mpi = _get_option(config_settings, 'MPI', '0')
    print("** MPI:", mpi)
    addlist = ["mpi4py"] if mpi else []
    print("** Additional modules:", addlist)
    ret = _orig.get_requires_for_build_wheel(config_settings) + addlist
    print("** Leaving get_requires_for_build_wheel")
    return ret



"""
Original author of HarmonicAnalysis class
Jaime Maria Coello De Portugal - Martinez Vazquez


Written while at CERN circa 2016
This is a reduced version keeping only the key components.
Full code can be found at: https://github.com/pylhc/harpy

"""

import numpy as np
from typing import Optional, Tuple

PI2I = 2 * np.pi * complex(0, 1)

ZERO_PAD_DEF = False
HANN_DEF = False


class HarmonicAnalysis(object):

    def __init__(self, samples, zero_pad=ZERO_PAD_DEF, hann=HANN_DEF):
        self._samples = samples
        self._compute_orbit()
        if zero_pad:
            self._pad_signal()
        self._length = len(self._samples)
        self._int_range = np.arange(self._length)
        self._hann_window = None
        if hann:
            self._hann_window = np.hanning(self._length)

    def laskar_method(self, num_harmonics):
        samples = self._samples[:]  # Copy the samples array.
        n = self._length
        coefficients = []
        frequencies = []
        for _ in range(num_harmonics):
            # Compute this harmonic frequency and coefficient.
            dft_data = HarmonicAnalysis._fft(samples)
            frequency = self._jacobsen(dft_data)
            coefficient = HarmonicAnalysis._compute_coef(
                samples,
                frequency * n
            ) / n

            # Store frequency and amplitude
            coefficients.append(coefficient)
            frequencies.append(frequency)

            # Subtract the found pure tune from the signal
            new_signal = coefficient * np.exp(PI2I
                                              * frequency * self._int_range)
            samples = samples - new_signal

        coefficients, frequencies = zip(*sorted(zip(coefficients,
                                                    frequencies),
                                        key=lambda tuple: np.abs(tuple[0]),
                                        reverse=True))

        return frequencies, coefficients

    def get_signal(self):
        if self._hann_window is not None:
            return self._samples * self._hann_window
        else:
            return self._samples

    def get_coefficient_for_freq(self, freq):
        return self._compute_coef(self._samples,
                                  freq * self._length) / self._length

    def _pad_signal(self):
        """
        Pads the signal with zeros to a "good" FFT size.
        """
        length = len(self._samples)
        # TODO Think proper pad size
        pad_length = (1 << (length - 1).bit_length()) - length
        # pad_length = 6600 - length
        self._samples = np.pad(
            self._samples,
            (0, pad_length),
            'constant'
        )
        # self._samples = self._samples[:6000]

    def _jacobsen(self, dft_values):
        """
        This method interpolates the real frequency of the
        signal using the three highest peaks in the FFT.
        """
        k = np.argmax(np.abs(dft_values))
        n = self._length
        r = dft_values
        delta = np.tan(np.pi / n) / (np.pi / n)
        kp = (k + 1) % n
        km = (k - 1) % n
        delta = delta * np.real((r[km] - r[kp]) / (2 * r[k] - r[km] - r[kp]))
        return (k + delta) / n

    @staticmethod
    def _compute_coef_simple(samples, kprime):
        """
        Computes the coefficient of the Discrete Time Fourier
        Transform corresponding to the given frequency (kprime).
        """
        n = len(samples)
        freq = kprime / n
        exponents = np.exp(-PI2I * freq * np.arange(n))
        coef = np.sum(exponents * samples)
        return coef

    @staticmethod
    def _compute_coef_goertzel(samples, kprime):
        """
        Computes the coefficient of the Discrete Time Fourier
        Transform corresponding to the given frequency (kprime).
        This function is faster than the previous one if compiled
        with Numba.
        """
        n = len(samples)
        a = 2 * np.pi * (kprime / n)
        b = 2 * np.cos(a)
        c = np.exp(-complex(0, 1) * a)
        d = np.exp(-complex(0, 1) *
                   ((2 * np.pi * kprime) / n) *
                   (n - 1))
        s0 = 0.
        s1 = 0.
        s2 = 0.
        for i in range(n - 1):
            s0 = samples[i] + b * s1 - s2
            s2 = s1
            s1 = s0
        s0 = samples[n - 1] + b * s1 - s2
        y = s0 - s1 * c
        return y * d

    def _compute_orbit(self):
        self.closed_orbit = np.mean(self._samples)
        self.closed_orbit_rms = np.std(self._samples)
        self.peak_to_peak = np.max(self._samples) - np.min(self._samples)

    @staticmethod
    def _conditional_import_compute_coef():
        """
        Checks if Numba is installed.
        If it is, it sets the compiled goertzel algorithm as the
        coefficient function to use. If it isn't, it uses the
        normal Numpy one.
        """
        try:
            from numba import jit
            # print("Using compiled Numba functions.")
            return jit(HarmonicAnalysis._compute_coef_goertzel,
                       nopython=True, nogil=True)
        except ImportError:
            # print("Numba not found, using numpy functions.")
            return HarmonicAnalysis._compute_coef_simple

    @staticmethod
    def _conditional_import_fft():
        """
        If SciPy is installed, it will set its fft as the one
        to use as it is slightly faster. Otherwise it will use
        the Numpy one.
        """
        try:
            from scipy.fftpack import fft as scipy_fft
            from scipy.fftpack import fftfreq as scipy_fftfreq
            fft = staticmethod(scipy_fft)
            fftfreq = staticmethod(scipy_fftfreq)
            # print("Scipy found, using scipy FFT.")
        except ImportError:
            from numpy.fft import fft as numpy_fft
            from numpy.fft import fftfreq as numpy_fftfreq
            fft = staticmethod(numpy_fft)
            fftfreq = staticmethod(numpy_fftfreq)
            # print("Scipy not found, using numpy FFT.")
        return fft, fftfreq


# Set up conditional functions on load
##############################################
HarmonicAnalysis._compute_coef = staticmethod(
    HarmonicAnalysis._conditional_import_compute_coef())
HarmonicAnalysis._fft, HarmonicAnalysis._fftfreq = \
    HarmonicAnalysis._conditional_import_fft()
##############################################


def get_spectrum_harmonic(cent, method: str = 'laskar',
                          num_harmonics: int = 20,
                          hann: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    """Frequency analysis of beam motion

    Parameters:
        cent:           Centroid motions of the particle
        method:         ``'laskar'`` or ``'fft'``. Default: ``'laskar'``
        num_harmonics:  Number of harmonic components to compute (before mask
          applied)
        hann:           Turn on Hanning window. Default: :py:obj:`False`

    Returns:
        frequency (ndarray): (num_harmonics,) array of frequencies
        amplitude (ndarray): (num_harmonics,) array of amplitudes
    """
    ha = HarmonicAnalysis(cent, hann=hann)

    if method == 'laskar':
        ha_tune, ha_amp = ha.laskar_method(num_harmonics=num_harmonics)
    elif method == 'fft':
        signal = ha.get_signal()
        fft = ha._fft(signal)
        ha_tune, ha_amp = ha._fftfreq(len(fft)), np.abs(fft)
    else:
        raise ValueError('The method ' + method + ' is undefined')

    ha_amp = np.abs(np.array(ha_amp))
    ha_tune = np.array(ha_tune)
    return ha_tune, ha_amp


def get_tunes_harmonic(cents, method: str = 'laskar',
                       num_harmonics: int = 20, hann: bool = False,
                       fmin: float = 0, fmax: float = 1) -> np.ndarray:
    """Computes tunes from harmonic analysis

    Parameters:
        cents:          Centroid motions of the particle
        method:         ``'laskar'`` or ``'fft'``. Default: ``'laskar'``
        num_harmonics:  Number of harmonic components to compute (before mask
          applied)
        fmin:           Lower bound for tune
        fmax:           Upper bound for tune
        hann:           Turn on Hanning window. Default: :py:obj:`False`

    Returns:
        tunes (ndarray):    numpy array of length len(cents), max of the
          spectrum within [fmin fmax]
    """
    def get_max_spectrum(freq, amp, fmin, fmax):
        msk = np.logical_and(freq >= fmin, freq <= fmax)
        amp = amp[msk]
        freq = freq[msk]
        tune = freq[np.argmax(amp)]
        return tune

    cents = np.array(cents)
    if cents.ndim > 1:
        npart = cents.shape[0]
    else:
        cents = [cents]
        npart = 1
    tunes = np.zeros(npart)

    for i in range(npart):
        freq, amp = get_spectrum_harmonic(cents[i],
                                          num_harmonics=num_harmonics,
                                          method=method,
                                          hann=hann)
        tunes[i] = get_max_spectrum(freq, amp, fmin, fmax)

    return tunes



from ..lattice import Lattice, get_rf_frequency, check_radiation, get_s_pos
from ..lattice import DConstant
from ..constants import clight
from ..tracking import lattice_pass
from .orbit import find_orbit4
import numpy
import functools
from typing import Optional

__all__ = ['frequency_control', 'get_mcf', 'get_slip_factor',
           'get_revolution_frequency', 'set_rf_frequency']


def frequency_control(func):
    """Function to be used as decorator for ``func(ring, *args, **kwargs)``

    If ``ring.radiation`` is :py:obj:`True` **and** ``dp``, ``dct`` or ``df``
    is specified in ``kwargs``, make a copy of ``ring`` with a modified
    RF frequency, remove ``dp``, ``dct`` or ``df`` from ``kwargs`` and call
    ``func`` with the modified ``ring``.

    If ``ring.radiation`` is :py:obj:`False` **or** no ``dp``, ``dct`` or
    ``df`` is specified in ``kwargs``, ``func`` is called unchanged.

    Examples::

        @frequency_control
        def func(ring, *args, dp=None, dct=None, **kwargs):
            pass
    """
    @functools.wraps(func)
    def wrapper(ring, *args, **kwargs):
        if ring.radiation:
            momargs = {}
            for key in ['dp', 'dct', 'df']:
                v = kwargs.pop(key, None)
                if v is not None:
                    momargs[key] = v
            if len(momargs) > 0:
                ring = set_rf_frequency(ring, **momargs, copy=True)
        return func(ring, *args, **kwargs)
    return wrapper


@check_radiation(False)
def get_mcf(ring: Lattice, dp: Optional[float] = 0.0,
            keep_lattice: bool = False, **kwargs) -> float:
    r"""Compute the momentum compaction factor :math:`\alpha`

    Parameters:
        ring:           Lattice description (radiation must be OFF)
        dp:             Momentum deviation. Defaults to :py:obj:`None`
        keep_lattice:   Assume no lattice change since the previous tracking.
          Default: :py:obj:`False`

    Keyword Args:
        DPStep (Optional[float]):       Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`

    Returns:
        mcf (float):    Momentum compaction factor :math:`\alpha`
    """
    dp_step = kwargs.pop('DPStep', DConstant.DPStep)
    fp_a, _ = find_orbit4(ring, dp=dp - 0.5*dp_step, keep_lattice=keep_lattice)
    fp_b, _ = find_orbit4(ring, dp=dp + 0.5*dp_step, keep_lattice=True)
    fp = numpy.stack((fp_a, fp_b),
                     axis=0).T  # generate a Fortran contiguous array
    b = numpy.squeeze(lattice_pass(ring, fp, keep_lattice=True), axis=(2, 3))
    ring_length = get_s_pos(ring, len(ring))
    alphac = (b[5, 1] - b[5, 0]) / dp_step / ring_length[0]
    return alphac


def get_slip_factor(ring: Lattice, **kwargs) -> float:
    r"""Compute the slip factor :math:`\eta`

    Parameters:
        ring:           Lattice description (radiation must be OFF)

    Keyword Args:
        dp (Optional[float]):       Momentum deviation
        DPStep (Optional[float]):       Momentum step size.
          Default: :py:data:`DConstant.DPStep <.DConstant>`

    Returns:
        eta (float):    Slip factor :math:`\eta`
    """
    gamma = ring.gamma
    etac = (1.0/gamma/gamma - get_mcf(ring, **kwargs))
    return etac


def get_revolution_frequency(ring,
                             dp: Optional[float] = None,
                             dct: Optional[float] = None) -> float:
    """Compute the revolution frequency of the full ring [Hz]

    Parameters:
        ring:       Lattice description
        dp:             Momentum deviation. Defaults to :py:obj:`None`
        dct:            Path lengthening. Defaults to :py:obj:`None`

    Returns:
        frev:       Revolution frequency [Hz]
    """
    lcell = ring.get_s_pos(len(ring))[0]
    frev = ring.beta * clight / lcell / ring.periodicity
    if dct is not None:
        frev *= lcell / (lcell + dct)
    elif dp is not None:
        # Find the path lengthening for dp
        rnorad = ring.radiation_off(copy=True) if ring.radiation else ring
        orbit = lattice_pass(rnorad, rnorad.find_orbit4(dp=dp)[0])
        dct = numpy.squeeze(orbit)[5]
        frev *= lcell / (lcell + dct)
    return frev


def set_rf_frequency(ring, frequency=None, dp=None, dct=None, **kwargs):
    """Set the RF frequency

    Parameters:
        ring:           Lattice description
        frequency:      RF frequency [Hz]. Default: nominal frequency.
        dp:             Momentum deviation. Defaults to :py:obj:`None`
        dct:            Path lengthening. Defaults to :py:obj:`None`

    Keyword Args:
        cavpts (Optional[Refpts]):  If :py:obj:`None`, look for ring.cavpts, or
          otherwise take all cavities.
        array (Optional[bool]):     If :py:obj:`False` (default), ``frequency``
          is applied to the selected cavities with the lowest frequency. The
          frequency of all the other selected cavities is scaled by the same
          ratio.

          If :py:obj:`True`, directly apply ``frequency`` to the selected
          cavities. The value must be broadcastable to the number of cavities.
        copy (Optional[bool]):     If :py:obj:`True`, returns a shallow copy of
          ``ring`` with new cavity elements. Otherwise (default), modify
          ``ring`` in-place
    """
    if frequency is None:
        frequency = ring.get_revolution_frequency(dp=dp, dct=dct) \
                    * ring.harmonic_number
    return ring.set_cavity(Frequency=frequency, **kwargs)


Lattice.mcf = property(get_mcf, doc="Momentum compaction factor")
Lattice.slip_factor = property(get_slip_factor, doc="Slip factor")
Lattice.get_revolution_frequency = get_revolution_frequency
Lattice.get_mcf = get_mcf
Lattice.get_slip_factor = get_slip_factor
Lattice.set_rf_frequency = set_rf_frequency
Lattice.rf_frequency = property(get_rf_frequency, set_rf_frequency,
    doc="Fundamental RF frequency [Hz]. The special value "
        ":py:class:`Frf.NOMINAL <.Frf>` means nominal frequency.")



import at
import numpy
import matlab
from numpy.testing import assert_allclose as assert_close
import pytest


@pytest.mark.parametrize('lattices',
                         [pytest.lazy_fixture('dba'),
                          pytest.lazy_fixture('hmba'),
                          pytest.lazy_fixture('hmba_cav'),
                          pytest.lazy_fixture('hmba_rad')])
def test_atpass(engine, lattices):
    py_lattice, ml_lattice, _ = lattices
    xy_step = 1.e-8
    scaling = [xy_step, xy_step, xy_step, xy_step, xy_step, xy_step]

    ml_rin = engine.diag(matlab.double(scaling))
    ml_rout = engine.atpass(ml_lattice, ml_rin, 1, 1)

    py_rin = numpy.asfortranarray(numpy.diag(scaling))
    at.lattice_pass(py_lattice, py_rin, 1)

    assert_close(py_rin, ml_rout, rtol=0, atol=1.e-30)


@pytest.mark.parametrize('lattices',
                         [pytest.lazy_fixture('dba'),
                          pytest.lazy_fixture('hmba'),
                          pytest.lazy_fixture('hmba_cav'),
                          pytest.lazy_fixture('hmba_rad')])
def test_linepass(engine, lattices):
    py_lattice, ml_lattice, _ = lattices
    xy_step = 1.e-8
    scaling = [xy_step, xy_step, xy_step, xy_step, xy_step, xy_step]

    ml_rin = engine.diag(matlab.double(scaling))
    ml_rout = numpy.array(engine.linepass(ml_lattice, ml_rin))

    py_rin = numpy.asfortranarray(numpy.diag(scaling))
    py_rout = numpy.squeeze(at.lattice_pass(py_lattice, py_rin,
                                            refpts=len(py_lattice)))

    assert_close(py_rout, ml_rout, rtol=0, atol=1.e-30)



"""
Helper functions for working with AT lattices.

A lattice as understood by pyAT is any sequence of elements.  These functions
are useful for working with these sequences.
"""
import sys
from .options import DConstant
from .particle_object import Particle
from .elements import *
from .utils import *
from .lattice_object import *
from .cavity_access import *
# Define the module "lattice.constants" for backward compatibility
# noinspection PyUnresolvedReferences
from .. import constants
sys.modules['at.lattice.constants'] = sys.modules['at.constants']



import numpy
import pytest
from at.physics import get_radiation_integrals


@pytest.mark.parametrize('lattices',
                         [pytest.lazy_fixture('hmba')])
def test_radiation_integrals(engine, lattices):
    py_lattice, ml_lattice, _ = lattices

    # Python call
    py_integrals = get_radiation_integrals(py_lattice)
    # Matlab call
    results = engine.pyproxy('ringpara', ml_lattice)
    ml_integrals = numpy.squeeze(results['integrals'])
    # Comparison
    numpy.testing.assert_almost_equal(py_integrals, ml_integrals[:5])



"""
Tracking functions
"""

# noinspection PyUnresolvedReferences
from .atpass import isopenmp, ismpi
from .patpass import patpass
from .track import *
from .particles import *
# noinspection PyProtectedMember
from ..lattice.options import _Dst

if ismpi():
    # if AT is compiled with mpicc this is required
    # noinspection PyUnresolvedReferences
    from mpi4py import MPI

_Dst.openmp = property(lambda self: isopenmp())
_Dst.mpi = property(lambda self: ismpi())
